# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import subprocess

from abc import abstractmethod, ABCMeta

from autopilot.exceptions import ProcessSearchError
from psutil import NoSuchProcess
from retrying import retry

from ubuntu_system_tests.helpers import autopilot
from ubuntu_system_tests.helpers import processes
from ubuntu_system_tests.helpers.snap import (
    get_snap_revision,
    is_installed as is_snap_installed,
)
from ubuntu_system_tests.helpers.unity8.utils import (
    is_unity8_snap,
    launch_application_from_app_drawer,
    launch_application_from_launcher,
)
from ubuntu_system_tests.helpers.utils import is_discovery

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class Application:
    __metaclass__ = ABCMeta

    def __init__(self, app_name, app_id):
        """
        :param app_name: Display name for application.
        :param app_id: Application id.
        """
        self.app_name = app_name
        self.app_id = app_id
        self.proxy_object = None
        self.cmd_prefix = None

    def _get_cmd_prefix(self):
        """Return prefix to use with all ubuntu-app-launch commands.
        This will depend on whether unity8 is running from a snap
        session or not.
        """
        if self.cmd_prefix is None:
            self.cmd_prefix = 'unity8-session.' if is_unity8_snap() else ''
        return self.cmd_prefix

    @abstractmethod
    def _get_app_rev_id(self, app_id):
        """Return the application revision ID."""
        pass

    def _get_start_command(self):
        """Return command to start application."""
        return [
            self._get_cmd_prefix() + 'url-dispatcher',
            self._get_launch_params()]

    def _get_stop_command(self):
        """Return command to stop application."""
        id = self._get_app_rev_id()
        return [self._get_cmd_prefix() + 'ubuntu-app-stop', id]

    def _get_pid_command(self):
        """Return command to get application process ID."""
        id = self._get_app_rev_id()
        return [self._get_cmd_prefix() + 'ubuntu-app-pid', id]

    def stop(self):
        """Stop application."""
        cmd = self._get_stop_command()
        logger.info('application stop command: {}'.format(' '.join(cmd)))
        subprocess.check_call(cmd)

    def get_pid(self):
        """Return application process ID."""
        cmd = self._get_pid_command()
        logger.info('application pid command: {}'.format(' '.join(cmd)))
        try:
            out = subprocess.check_output(cmd).decode()
        except subprocess.CalledProcessError:
            pass
        else:
            for line in out.splitlines():
                if line.isdigit():
                    return line
        raise RuntimeError(
            'Process ID not found for app ID: {}'.format(self.app_id))

    def running(self):
        """Return boolean to indicate if application is running."""
        return subprocess.call(self._get_pid_command()) == 0

    def launch(self):
        """Launch application using command."""
        cmd = self._get_start_command()
        logger.info('application launch command: {}'.format(' '.join(cmd)))
        subprocess.check_call(cmd)

    def launch_from_app_drawer(self, return_proxy=True):
        """
        Launch app from app drawer.
        """
        launch_application_from_app_drawer(self)
        if return_proxy:
            return self.get_proxy_object()

    def launch_from_launcher(self, return_proxy=True):
        """Drag out the launcher and tap the application icon to start it."""
        launch_application_from_launcher(self.app_id)
        if return_proxy:
            return self.get_proxy_object()

    @retry(stop_max_delay=3000,
           wait_fixed=1000,
           retry_on_exception=lambda exception: (
               isinstance(exception, ProcessSearchError) or
               isinstance(exception, processes.PidNotFoundError) or
               isinstance(exception, RuntimeError) or
               isinstance(exception, NoSuchProcess)))
    def get_proxy_object(self):
        if self.proxy_object is None:
            self.proxy_object = autopilot.get_proxy_object(int(self.get_pid()))
        return self.proxy_object

    @abstractmethod
    def get_main_view(self):
        pass

    def is_installed(self):
        if is_discovery():
            # This is a special case where test cases are being discovered
            # during a list command.
            # In this case always return True to ensure all tests can be
            # listed on the host machine.
            return True
        return self._is_installed()

    @abstractmethod
    def _is_installed(self, app_id):
        """Return boolean indicating if app is installed."""
        pass


class Snap(Application):

    app_type = 'snap'

    def __init__(self, app_name, app_id):
        super().__init__(app_name, app_id)
        self.app_rev_id = None

    def _get_launch_params(self):
        """Return required parameters for app launch command."""
        return 'appid://{n}/{n}/current-user-version'.format(n=self.app_id)

    def _get_app_rev_id(self):
        """Return the application revision ID."""
        if self.app_rev_id is None:
            rev = get_snap_revision(self.app_id)
            self.app_rev_id = '{n}_{n}_{r}'.format(n=self.app_id, r=rev)
        return self.app_rev_id

    def _is_installed(self):
        """Return boolean indicating if app is installed."""
        return is_snap_installed(self.app_id)


class Deb(Application):

    app_type = 'deb'

    def __init__(self, app_name, app_id, package_id=None):
        super().__init__(app_name, app_id)
        self.package_id = package_id or app_id

    def _get_launch_params(self):
        """Return required parameters for app launch command."""
        return 'application:///{n}.desktop'.format(n=self.app_id)

    def _get_app_rev_id(self):
        """Return the application revision ID."""
        return self.app_id

    def _is_installed(self):
        """Return boolean indicating if app is installed."""
        output = subprocess.check_output(
            ['apt-cache', 'policy', self.package_id])
        return len(output) > 0 and 'Installed: (none)' not in output.decode()
