# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os

from autopilot import (
    platform,
    input
)

from ubuntu_system_tests.helpers.init_systems import get_init_system
from ubuntu_system_tests.helpers.ubuntu_keyboard.job import MaliitServerJob
from ubuntu_system_tests.helpers.ubuntu_keyboard.settings import (
    configure_osk_settings,
    restart_maliit_with_testability
)
from ubuntu_system_tests.helpers.utils import Singleton


class InputManager(metaclass=Singleton):
    """Class to provide a centralized interface for getting and releasing input
     devices.
     """

    def __init__(self):
        self.__pointer = None
        self.__keyboard = None
        self.__osk = False

    @property
    def pointer(self):
        return self.__get_pointing_device()

    @property
    def keyboard(self):
        return self.__get_keyboard()

    def set_pointing_device(self):
        self.__pointer = self.__create_pointing_device()

    def set_keyboard(self):
        self.__keyboard = self.__create_keyboard()

    def release_pointing_device(self):
        self.__pointer = None

    def release_keyboard(self):
        self.__keyboard = None

    def is_osk(self):
        return self.__osk

    def __get_pointing_device(self):
        """Get a pointing device instance.
        It will return the same instance until the function for releasing
        it is called: release_pointing_device.
        :return a pointing device instance.
        """
        if self.__pointer is None:
            self.set_pointing_device()
        return self.__pointer

    def __get_keyboard(self):
        """Get a keyboard instance.
        It will return the same instance until function for releasing
        it is called: release_keyboard.
        :return a keyboard instance.
        """
        if self.__keyboard is None:
            self.set_keyboard()
        return self.__keyboard

    def __create_pointing_device(self):
        """Return the pointing device depending on the platform.
        If the platform is `Desktop`, the pointing device will be a `Mouse`.
        If not, the pointing device will be `Touch`.
        """
        if platform.model() == 'Desktop':
            input_device_class = input.Mouse
        else:
            input_device_class = input.Touch
        return input.Pointer(device=input_device_class.create())

    def __create_keyboard(self):
        """Return the keyboard device."""
        if get_init_system().is_running(MaliitServerJob()) and os.environ.get(
                'UITK_USE_MALIIT'):
            configure_osk_settings()
            restart_maliit_with_testability()
            self.__osk = True
            return input.Keyboard.create('OSK')
        self.__osk = False
        return input.Keyboard.create()


input_manager = InputManager()
