# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from autopilot import logging as autopilot_logging
from ubuntu_system_tests.helpers.autopilot import validate_dbus_object
from ubuntu_system_tests.helpers.input_manager import input_manager
from ubuntu_system_tests.helpers.ubuntuuitools.cpo.common import (
    CustomProxyObjectBase,
)
from ubuntu_system_tests.helpers.ubuntuuitools.cpo.qquicklistview import (
    QQuickListView,
)
from ubuntu_system_tests.helpers.ubuntuuitools.cpo.textfield import (
    TextField,
)
from ubuntu_system_tests.helpers.unity8.utils import UNITY8_PATH_ROOT

logger = logging.getLogger(__name__)


class Drawer(CustomProxyObjectBase):
    """Helper for the app drawer."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, UNITY8_PATH_ROOT, b'Drawer', objectName='drawer')

    def _get_search_textfield(self):
        return self.select_single(TextField, objectName='searchField')

    @autopilot_logging.log_action(logger.info)
    def open(self):
        """Open the app drawer if its not already displayed."""
        if not self.visible:
            input_manager.keyboard.press_and_release('LeftMeta+a')
            self.visible.wait_for(True)
            self.wait_until_not_moving()

    @autopilot_logging.log_action(logger.info)
    def close(self):
        """Close the app drawer if it is being displayed."""
        if self.visible:
            input_manager.keyboard.press_and_release('LeftMeta')
            self.visible.wait_for(False)
            self.wait_until_not_moving()

    @autopilot_logging.log_action(logger.info)
    def search(self, query):
        """Enter search query into search field."""
        search = self._get_search_textfield()
        if search.displayText != query:
            search.clear()
            search.write(query)

    @autopilot_logging.log_action(logger.info)
    def clear_search(self):
        """Clear current search query from search field."""
        search = self._get_search_textfield()
        if search.displayText:
            search.clear()

    @autopilot_logging.log_action(logger.info)
    def select_application(self, application):
        """Scroll application into view and select it."""
        list_view = self.select_single(
            DrawerListView, objectName='drawerItemList')
        list_view.select_application(application)

    @autopilot_logging.log_action(logger.info)
    def open_application(self, application):
        """Search for an application and select if from results."""
        self.search(application.app_name)
        self.select_application(application)


class DrawerListView(QQuickListView):
    """Helper to scroll and select displayed applications."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, UNITY8_PATH_ROOT, b'DrawerListView',
            objectName='drawerItemList')

    def select_application(self, application):
        """Search for an application in list view and select it."""
        if application.app_type == 'snap':
            name = 'drawerItem_{n}_{n}'
        elif application.app_type == 'deb':
            name = 'drawerItem_{n}'
        else:
            raise NameError(
                '{} is not supported application '
                'type'.format(application.app_type))
        obj_name = name.format(n=application.app_id)
        self.click_element(obj_name)
