# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014-2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import psutil
import subprocess

from autopilot import exceptions
from autopilot.platform import model
from dbus import DBusException
from retrying import retry
from ubuntu_system_tests.helpers import (
    context,
    timer
)
from ubuntu_system_tests.helpers import wait_until
from ubuntu_system_tests.helpers.autopilot import get_proxy_object
from ubuntu_system_tests.helpers.init_systems import get_init_system
from ubuntu_system_tests.helpers.processes import (
    is_snap_process,
    is_testability_enabled_for_process,
    get_process_id,
    PidNotFoundError)
from ubuntu_system_tests.helpers.testbed import run_command_with_sudo
from ubuntu_system_tests.helpers.unity8 import greeter
from ubuntu_system_tests.helpers.unity8.job import Unity8Job
from ubuntu_system_tests.helpers.user import (
    get_device_username,
    get_device_user_id
)
from ubuntu_system_tests.tests.data import DESKTOP_PLATFORM_NAME

UNITY8_PATH_ROOT = b'unity8'
UNITY8_DASH_PATH_ROOT = b'unity8-dash'
APP_WIN_NAME = 'Scopes'
WIZARD_FILE = '.config/ubuntu-system-settings/wizard-has-run'
UNITY8_SNAP_COMMON_PATH = 'snap/unity8-session/common'
XENIAL = '16.04'


def is_unity8_snap():
    """Return True if unity8 process is running from snap, False otherwise."""
    return is_snap_process(get_process_id('unity8'))


def lock_unity():
    """Helper function that attempts to lock unity greeter."""
    greeter.show_greeter_with_dbus()
    greeter.wait_for_greeter()


def unlock_unity():
    """Helper function that attempts to unlock the unity greeter."""
    greeter.wait_for_greeter()
    greeter.hide_greeter_with_dbus()
    greeter.wait_for_greeter_gone()


def get_dash():
    """Return the Unity8 Dash autopilot custom proxy object."""
    from ubuntu_system_tests.helpers.unity8.dash import Dash
    dash_proxy = get_proxy_object(get_process_id('unity8-dash'))
    return dash_proxy.select_single(Dash)


def launch_application_from_launcher(app_name):
    """
    Launch an app from the launcher.

    :param app_name: Name of the displayed application.

    """
    unity = context.shared.unity
    timer.start(timer.LAUNCH_ID)
    unity.main_window.launch_application(app_name)


def launch_application_from_app_drawer(application):
    """
    Launch an app from the app drawer.

    :param application: Application derived object for required
                        application.

    """
    unity = context.shared.unity
    timer.start(timer.LAUNCH_ID)
    unity.main_window.launch_application_from_app_drawer(application)


def close_app(app_name):
    """Close an app using task switcher and check process is stopped."""
    unity = context.shared.unity

    # Get the windows name
    stage = unity.main_window.swipe_to_show_app_switcher()
    app_windows = stage.get_app_window_names()
    if app_name not in app_windows:
        raise RuntimeError(
            'Window {} no included in windows list {}'.format(
                app_name, app_windows
            ))

    # Close the windows and assert that it is closed
    stage.close_app(app_name)
    app_windows = stage.get_app_window_names()
    if app_name in app_windows:
        raise RuntimeError(
            'Window {} not closed correctly'.format(app_name))


def close_all_apps():
    """ Close all the apps using task switcher """
    unity = context.shared.unity
    stage = unity.main_window.swipe_to_show_app_switcher()
    app_names = stage.get_app_window_names()
    for app_name in app_names:
        if not app_name == APP_WIN_NAME:
            close_app(app_name)


def ensure_unity_in_required_state(test_id):
    unity8 = Unity8Job()
    if test_id.startswith('unity8'):
        get_init_system().ensure_stopped(unity8)
    else:
        unity8.ensure_unity_running_and_greeter_hidden()


def _is_unity_shared_context_setup():
    try:
        context.shared.unity_pid
        context.shared.unity
        context.shared.dash
    except AttributeError:
        return False
    else:
        return True


def _initialize_unity_shared_context_and_unlock(init_system, unity8_job):
    context.shared.restart_unity = False
    context.shared.unity_pid = init_system.get_pid(unity8_job)
    context.shared.unity = get_proxy_object(pid=context.shared.unity_pid)
    if model() != DESKTOP_PLATFORM_NAME:
        unity8_job.wait_for_unity_ready()
        context.shared.unity.wait_for_greeter_content()
        context.shared.unity.unlock()
    context.shared.dash = get_dash()


# Retry added until fix for bug 1427946 in QtMir is landed
@retry(stop_max_attempt_number=5,
       wait_fixed=2000,
       retry_on_exception=lambda exception: (
           isinstance(exception, psutil.NoSuchProcess) or
           isinstance(exception, ValueError) or
           isinstance(exception, DBusException) or
           isinstance(exception, TimeoutError) or
           isinstance(exception, PidNotFoundError) or
           isinstance(exception, exceptions.ProcessSearchError) or
           isinstance(exception, IndexError) or
           isinstance(exception, exceptions.BackendException)))
def setup_unity(restart=False):
    from ubuntu_system_tests.helpers.unity8.shell import Unity8  # NOQA
    init_system = get_init_system()
    unity8 = Unity8Job()
    restart = restart or getattr(context.shared, 'restart_unity', False)
    if restart or not is_testability_enabled_for_process(
            init_system.get_pid(unity8)):
        init_system.ensure_stopped(unity8)
        init_system.start(unity8)
        wait_until(init_system.is_running, unity8)
        _initialize_unity_shared_context_and_unlock(init_system, unity8)
    elif not _is_unity_shared_context_setup():
        _initialize_unity_shared_context_and_unlock(init_system, unity8)


def is_settings_wizard_enabled():
    return not os.path.exists(get_wizard_file_full_path())


def set_settings_wizard(enabled):
    """Set settings wizard from running on first boot.
    :return: True if wizard config value has changed, False otherwise.
    """
    if enabled:
        return enable_settings_wizard()
    else:
        return disable_settings_wizard()


def enable_settings_wizard():
    """Enable settings wizard from running on first boot.
    :return: True if wizard config value has changed, False otherwise.
    """
    wizard_file = get_wizard_file_full_path()
    if os.path.exists(wizard_file):
        os.remove(wizard_file)
        context.shared.restart_unity = True
        return True
    return False


def disable_settings_wizard():
    """Disable settings wizard from running on first boot.
    :return: True if wizard config value has changed, False otherwise.
    """
    wizard_file = get_wizard_file_full_path()
    if not os.path.isfile(wizard_file):
        wizard_dir = os.path.dirname(wizard_file)
        if not os.path.isdir(wizard_dir):
            os.makedirs(wizard_dir)
        open(wizard_file, 'a').close()
        context.shared.restart_unity = True
        return True
    return False


def get_wizard_file_full_path():
    # Get path of wizard-has-run file depending on whether snap is being
    # used or not.
    base_path = '/home/{}'.format(get_device_username())
    snap_base_path = os.path.join(base_path, UNITY8_SNAP_COMMON_PATH)
    if os.path.exists(snap_base_path):
        base_path = snap_base_path
    wizard_file = os.path.join(base_path, WIZARD_FILE)
    return wizard_file


def is_edge_intro_enabled():
    edge_intro_command = build_edge_intro_command('Get')
    return 'boolean true' in run_edge_intro_command(
        edge_intro_command).decode()


def disable_edges_intro():
    """Disable edges intro from running on first boot.
    :return: True if edge intro config value has changed, False otherwise.
    """
    return set_edge_intro(enable=False)


def set_edge_intro(enable=True):
    """Set edges intro from running on first boot: True to enable it
    Only execute the command when the given value is different
    from the current one.
    :return: True if edge intro config value has changed, False otherwise.
    """
    if enable != is_edge_intro_enabled():
        edge_intro_command = build_edge_intro_command(
            'Set', 'variant:boolean:{}'.format(str(enable).lower()))
        run_edge_intro_command(edge_intro_command)
        context.shared.restart_unity = True
        return True
    return False


def build_edge_intro_command(operation, *args):
    """Build and return the command to query or update demo edges feature
    :param operation: a string that can be 'Set' to update or 'Get' to query
    :param args: extra arguments used in case of a 'Set' operation.
    For example: variant:boolean:true
    :return: the edge intro command ready to be executed
    """
    user_path = '/org/freedesktop/Accounts/User{}'.format(get_device_user_id())
    cmd = [
        'dbus-send', '--system', '--print-reply',
        '--dest=org.freedesktop.Accounts', user_path,
        'org.freedesktop.DBus.Properties.{}'.format(operation.capitalize()),
        'string:com.canonical.unity.AccountsService',
        'string:demo-edges']
    for arg in args:
        cmd.append(arg)
    return cmd


def run_edge_intro_command(cmd):
    # TODO: remove conditional when support for vivid ends
    if get_release() < XENIAL:
        # For vivid ubuntu-touch, this needs to be run as root
        # so no user args required
        return run_command_with_sudo(' '.join(cmd))
    else:
        # run command as normal user to disable edges demo
        return subprocess.check_output(cmd)


def get_release():
    """Return release name of running system."""
    return subprocess.check_output(
        ['lsb_release', '-r', '-s']).decode().strip()
