#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014-2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import subprocess
import sys

from argparse import ArgumentParser
from collections import namedtuple

from ubuntu_system_tests.host import (
    commands as cmds,
    ppa,
    qemu,
    qemu_utils,
    results,
    targets,
)
from ubuntu_system_tests.common.utils import clean_dir
from ubuntu_system_tests.common import config

from ubuntu_system_tests.host.practitest import PractitestSession

logger = logging.getLogger(__name__)

FILTER_ID_TEST = 'test'
FILTER_ID_FILTER = 'filter'
FILTER_ALIAS = {
    'sanity': 'filter:63805',
    'regression': 'filter:117776',
    'stable': 'filter:126345'
}
FIELD_NAME = 'name'
FIELD_VALUE = 'value'
FIELD_EXTERNAL_ID = 'External ID'

Option = namedtuple('Option', ['args', 'kwargs'])
OPTION_APPEND_RESULTS = Option(
    ['--append-results'],
    {'action': 'store_true',
     'help': 'Append test results to existing results in output directory.',
     'default': False})
OPTION_CACHE_DEPS = Option(
    ['--cache-deps', '-d'],
    {'action': 'store_true',
     'help': 'Cache dependencies on the device so they don\'t have to be '
     'downloaded for every test run.',
     'default': False})
OPTION_CONFIG = Option(
    ['--config', '-c'],
    {'help': 'The config file to use for this run.',
     'default': 'ubuntu-system-tests.conf'})
OPTION_CONFIG_SECTION = Option(
    ['--config-section', '-e'],
    {'help': 'The section name to use in config file.',
     'default': config.KEY_DEFAULT})
OPTION_DIST_UPGRADE = Option(
    ['--dist-upgrade', '-p'],
    {'action': 'store_true',
     'help': 'Perform a dist-upgrade on the device to install latest '
     'available updates from archive.',
     'default': False})
OPTION_IP = Option(
    ['--ip'],
    {'help': 'IP address of target. This will override the value defined in '
     'config file.'})
OPTION_PORT = Option(
    ['--port'],
    {'help': 'Port number of target. This will override the value defined in '
     'config file.'})
OPTION_INSTALL_TESTS = Option(
    ['--install-tests', '-i'],
    {'action': 'store_true',
     'help': 'Install all dependencies permanently in the device instead '
     'of in the /tmp directory. This includes all production and test '
     'dependencies.',
     'default': False})
OPTION_NETWORK_READY = Option(
    ['--network-ready', '-n'],
    {'action': 'store_true',
     'help': 'Assume that the network connection is already setup.',
     'default': False})
OPTION_FILTERS = Option(
    ['--filters', '-f'],
    {'help': 'Comma seperated list of Practitest filter IDs to use to '
     'generate a list of additional tests to add to specified command. These '
     'could be the ID of a single test or ID of a filter that contains '
     'multiple tests. The filters should be specified as \'filter:<ID>\' or '
     '\'test:<ID>\' to distinguish between a test ID or filter ID. If no '
     'prefix is defined then it is assumed to be a filter. Alias filter names '
     'of {a} are also supported. The list of tests returned from Practitest '
     'will be used in addition to any tests specified in the '
     'suites parameter.'.format(a=', '.join(FILTER_ALIAS.keys()))})
OPTION_QEMU_HEADLESS = Option(
    ['--headless', '-l'],
    {'action': 'store_true',
     'help': 'Do not show UI when running QEMU.',
     'default': False})
OPTION_QEMU_IMG_PATH = Option(
    ['--img', '-m'],
    {'help': 'Path to local IMG file to use for QEMU image.'})
OPTION_QEMU_ISO_PATH = Option(
    ['--iso', '-i'],
    {'help': 'Path to local iso file to install into QEMU image.'})
OPTION_QEMU_NO_OVERLAY = Option(
    ['--no-overlay', '-o'],
    {'action': 'store_true',
     'help': 'For QEMU targets, run the QEMU without a temporary overlay '
             'image to allow persistent changes to be made to the base image. '
             'If this option is not specified then by default a QEMU will '
             'be launched with a temporary overlay image to prevent any '
             'persistent changes being made to the base image.',
     'default': False})
OPTION_LIST_SUITES = Option(
    ['suites'],
    {'nargs': '*',
     'help': 'Optional list of test suites to expand. All test suites will be '
     'listed to show test IDs for every test case in the suite.'})
OPTION_PPA_LIST = Option(
    ['--ppa', '-a'],
    {'help': 'Comma separated list of additional PPAs to install on device '
     'during setup. Standard format should be ppa:owner/name, but an alias '
     'of silo:<ID> can be used to lookup PPAs to add from CI train.'})
OPTION_RUN_SUITES = Option(
    ['suites'],
    {'nargs': '*',
     'help': 'The list of test suites to run. If left blank then all tests '
     'from ubuntu_system_tests.tests will be run.'})
OPTION_SETUP_SUITES = Option(
    ['suites'],
    {'nargs': '*',
     'help': 'The list of test suites to download or install dependencies '
     'for. If left blank then all tests from ubuntu_system_tests will be '
     'used. This option is used only if --cache-deps or --install-tests is '
     'enabled.'})
OPTION_SILENT = Option(
    ['--silent', '-s'],
    {'action': 'store_true',
     'help': 'Do not prompt for missing user config options, use default '
     'values instead.',
     'default': False})
OPTION_TOUCH_VISUALIZATION = Option(
    ['--touch-visualization', '-t'],
    {'action': 'store_true',
     'help': 'Enable visualization of touch points on screen.',
     'default': False})
OPTION_RUN_UPGRADE = Option(
    ['--upgrade'],
    {'help': 'Comma separated list of components to upgrade on device prior '
     'to running any tests. This is useful for QEMU targets when running with '
     'an overlay image to allow upgrades to be made temporarily during the '
     'test run without changing the base image.'})
OPTION_UPDATE_APT = Option(
    ['--update-apt', '-u'],
    {'action': 'store_true',
     'help': 'Update apt lists before installing the dependencies.',
     'default': False})
OPTION_VERBOSE = Option(
    ['--verbose', '-v'],
    {'action': 'store_true',
     'help': 'Provide verbose logging.',
     'default': False})

COMMON_ARGS = [
    OPTION_CONFIG,
    OPTION_CONFIG_SECTION,
    OPTION_SILENT,
]
COMMON_RUN_SETUP_LIST_ARGS = [
    OPTION_FILTERS,
]
COMMON_RUN_SETUP_ARGS = [
    OPTION_NETWORK_READY,
    OPTION_VERBOSE,
    OPTION_QEMU_IMG_PATH,
    OPTION_QEMU_HEADLESS,
    OPTION_IP,
    OPTION_PORT,
]
LAUNCH_ARGS = [
    OPTION_QEMU_IMG_PATH,
    OPTION_QEMU_NO_OVERLAY,
    OPTION_QEMU_HEADLESS,
    OPTION_VERBOSE,
]
LIST_ARGS = [
    OPTION_LIST_SUITES,
]
RUN_ARGS = [
    OPTION_RUN_SUITES,
    OPTION_TOUCH_VISUALIZATION,
    OPTION_QEMU_NO_OVERLAY,
    OPTION_APPEND_RESULTS,
    OPTION_RUN_UPGRADE,
]
SETUP_ARGS = [
    OPTION_CACHE_DEPS,
    OPTION_DIST_UPGRADE,
    OPTION_INSTALL_TESTS,
    OPTION_UPDATE_APT,
    OPTION_PPA_LIST,
    OPTION_SETUP_SUITES,
]
BUILD_QEMU_ARGS = [
    OPTION_QEMU_IMG_PATH,
    OPTION_QEMU_ISO_PATH,
    OPTION_QEMU_HEADLESS,
    OPTION_VERBOSE,
]

COMMAND_BUILD_QEMU = 'build-qemu'
COMMAND_LAUNCH = 'launch'
COMMAND_LIST = 'list'
COMMAND_RUN = 'run'
COMMAND_SETUP = 'setup'
COMMANDS = [
    COMMAND_BUILD_QEMU,
    COMMAND_LAUNCH,
    COMMAND_LIST,
    COMMAND_RUN,
    COMMAND_SETUP,
]

TRUE = 'true'


def main(argv=None):  # NOQA
    args = _parse_arguments(argv or sys.argv[1:])
    if args.command not in COMMANDS:
        logger.warning(
            'Command option must be either: {c}. Use -h option for '
            'further help.'.format(c=', '.join(COMMANDS)))
        return

    _validate_suites(args)
    if args.command == COMMAND_RUN:
        return run_system_tests(args)
    elif args.command == COMMAND_SETUP:
        return setup_system_tests(args)
    elif args.command == COMMAND_BUILD_QEMU:
        return build_qemu(args)
    elif args.command == COMMAND_LAUNCH:
        return launch(args)
    else:
        return list_system_tests(args)


def _validate_suites(args):
    """Validate that the suites parameter is correctly formatted."""
    if (hasattr(args, 'suites') and
            any(suite.find(' ') >= 0 for suite in args.suites)):
        # A suite contains a space character.
        # This means multiple test ids have been concatenated together as a
        # single string. Need to separate them into a list before proceeding.
        suites = []
        for suite in args.suites:
            suites.extend(suite.strip().split(' '))
        args.suites = suites


def _add_options(parser, options):
    """Add the specified list of options to the option parser object."""
    for option in options:
        parser.add_argument(*option.args, **option.kwargs)


def _parse_arguments(argv=None):
    """Parse the provided arguments and return in Namespace object."""
    parser = ArgumentParser(
        'ubuntu-system-tests',
        description='Autopilot system test suite for Ubuntu Touch devices. '
        'Allows devices to be setup for testing and then used to execute '
        'suites of autopilot tests from ubuntu-system-tests or other '
        'projects. For further help use -h option with required command.')
    subparsers = parser.add_subparsers(
        dest='command',
        help='Command options')
    build_qemu_cmd = subparsers.add_parser(
        COMMAND_BUILD_QEMU,
        help='Build a qemu image',
        description='Optionally download an ISO file or use an existing local '
        'ISO file to use for installation into a qemu image to use for '
        'testing.')
    _add_options(build_qemu_cmd, COMMON_ARGS + BUILD_QEMU_ARGS)
    launch_cmd = subparsers.add_parser(
        COMMAND_LAUNCH,
        help='Launch a qemu image and wait for it to terminate.',
        description='This is useful if using the ssh target to run tests '
        'when the qemu can be left running between test executions for '
        'further investigation or debugging.')
    _add_options(launch_cmd, COMMON_ARGS + LAUNCH_ARGS)
    list_cmd = subparsers.add_parser(
        COMMAND_LIST,
        help='List tests to run',
        description='Print a list of tests to run. This can include suites '
        'defined locally within ubuntu_system_tests or suites defined by '
        'Practitest filters.')
    _add_options(list_cmd, COMMON_ARGS + COMMON_RUN_SETUP_LIST_ARGS +
                 LIST_ARGS)
    run = subparsers.add_parser(
        COMMAND_RUN,
        help='Run tests on the device',
        description='Executes a list of tests on a connected device once it '
        'has been setup using the setup command. If no tests are specified '
        'to be run then all tests from ubuntu_system_tests.tests will be '
        'executed.')
    _add_options(run, COMMON_ARGS + COMMON_RUN_SETUP_LIST_ARGS +
                 COMMON_RUN_SETUP_ARGS + RUN_ARGS)
    setup = subparsers.add_parser(
        COMMAND_SETUP,
        help='Setup device ready for testing',
        description='Sets up the device ready for testing. This can include '
        'performing apt-get update and dist-upgrade operations. Also it is '
        'possible to install apt-cacher-ng to allow downloaded dependencies '
        'to be cached on the device so they don\'t need to be downloaded for '
        'every test run. If no arguments are specified then the default '
        'operation will be to perform apt-get update, dist-upgrade and '
        'install apt-cacher-ng. If any single option is specified then only '
        'this action will be performed.')
    _add_options(setup, COMMON_ARGS + COMMON_RUN_SETUP_LIST_ARGS +
                 COMMON_RUN_SETUP_ARGS + SETUP_ARGS)
    return parser.parse_args(args=argv)


def setup_system_tests(args):
    """Validate arguments and execute required setup commands on device."""
    if ((args.cache_deps or args.dist_upgrade or args.install_tests) and
            not args.update_apt):
        # An install operation is requested without update_apt, so set
        # update_apt to True.
        args.update_apt = True
    elif not (args.cache_deps or args.dist_upgrade or args.update_apt):
        # There were no optional args specified.
        # In this case apply the default actions, which are:
        # update_apt, cache_deps and dist_upgrade
        args.cache_deps = args.dist_upgrade = args.update_apt = True
    return run_setup(args)


def run_system_tests(args):
    """Run system tests in a device connected via USB and report the results.
    :param args: command line options.
    :returns: 0 when successful, other value when failed.

    """
    print_warnings()
    config_stack = _prepare_config(args)
    if not getattr(args, 'append_results', False):
        clean_output_directory(config_stack)
    if not add_tests_ids_for_filters(config_stack, args) and not args.suites:
        # In this case there were no tests returned from the specified filters
        # and no tests specified in the suites parameter. If execution were
        # to continue from this point it would run all tests under
        # ubuntu_system_tests, but as there has been an (invalid) filter
        # specified this is not the intended behaviour. So raise an exception
        # here to terminate execution.
        raise RuntimeError(
            'No test IDs returned for filters: {f}'.format(f=args.filters))
    return_code = run_autopkg_test(config_stack, args)

    if return_code <= 6:
        # Return codes higher than 6 mean that there was a problem with adt or
        # the test bed, not a failure in the tests. So there will be no file
        # with results.
        generate_user_report(config_stack)

    return return_code


def list_system_tests(args):
    """Print an ordered list of required test IDs with no duplicates."""
    config_stack = _prepare_config(args)
    clean_output_directory(config_stack)
    tests = []
    if not args.filters and not args.suites:
        # No list parameters specified, so set default
        args.suites = targets.TESTS_UST_DEFAULT
    if args.suites:
        tests.extend(_get_test_ids_for_suites(args.suites))
    if args.filters:
        tests.extend(_get_test_ids_for_filters(config_stack, args.filters))
    # Print sorted list with no duplicates
    print('\n'.join(sorted(list(set(tests)))))


def launch(args):
    """Launch qemu and wait for it to terminate."""
    config_stack = _prepare_config(args)
    # launch only supports qemu, so ensure the target type is correct
    config_stack.config['target_type'] = 'qemu'
    target = targets.get_target(config_stack, args)
    target.connect()
    ip = target.device_ip
    port = target.port
    user = target.username
    print('\ntarget running at: {i}:{p}'.format(i=ip, p=port))
    print('waiting for target to close.\n')
    print('useful commands:')
    print('open ssh:  ssh -p {p} {u}@{i}'.format(
        p=port, u=user, i=ip))
    print('copy to:   scp -r -P {p} <SRC> {u}@{i}:/home/{u}/<DST>'.format(
        p=port, u=user, i=ip))
    print('copy from: scp -r -P {p} {u}@{i}:/home/{u}/<SRC> <DST>'.format(
        p=port, u=user, i=ip))
    target.wait_for_stopped()


def build_qemu(args):
    """Build a qemu image using specified args."""
    config_stack = _prepare_config(args)
    clean_output_directory(config_stack)
    headless = (args.headless or
                config_stack.get('qemu_headless').lower() == TRUE)
    data_folder = os.path.realpath(config_stack.get('cache_data_path'))
    img_cache_dir = os.path.join(data_folder, 'img')
    img = os.path.realpath(os.path.expanduser(args.img or img_cache_dir))
    iso = args.iso or config_stack.get('qemu_iso_url')
    cpu = config_stack.get('qemu_cpu_count')
    ram = config_stack.get('qemu_ram_alloc')
    arch = config_stack.get('qemu_iso_arch')
    auto_download = config_stack.get('qemu_auto_iso_download').lower() == TRUE
    max_count = int(config_stack.get('qemu_max_cache_file_count'))
    img_disk_size = config_stack.get('qemu_disk_size')
    username = config_stack.get('device_username')
    password = config_stack.get('device_password')
    display_name = username

    qemu_utils.clean_cache_folder(data_folder, max_count)
    img_path = qemu_utils.get_local_img_file_path(img, max_count)
    iso_path = qemu_utils.get_local_iso_path(
        iso, arch, data_folder, auto_download)
    if not iso_path:
        # A download is required as no suitable iso could be found locally.
        # auto_download is already checked in get_local_iso_path, so its safe
        # to assume it is true here.
        iso_path = qemu_utils.download_iso(iso, arch, data_folder, max_count)

    qemu.build_qemu(
        img_path, iso_path, cpu, ram, img_disk_size, headless, args.verbose,
        username, password, display_name)
    # Cleanup the iso cache folder. Leave image file in place as this will
    # be used for further testing.
    qemu_utils.clean_iso_cache_folder(data_folder, max_count)
    print('QEMU image file created at: {}'.format(img_path))


def _get_test_ids_for_suites(suites):
    """Return a list of test IDs for specified list of suites."""
    tests = []
    for suite in suites:
        tests.extend(_list_autopilot_tests(suite))
    return tests


def _list_autopilot_tests(suite):
    """Return a list of autopilot test IDs for specified suite."""
    cmd = ['autopilot3', 'list', suite]
    try:
        out = subprocess.check_output(cmd, stderr=subprocess.DEVNULL).decode()
    except subprocess.CalledProcessError:
        raise ValueError('Invalid suite name: {s}'.format(s=suite))
    if 'ImportError:' in out:
        raise ImportError('Cannot import suite: {s}'.format(s=suite))
    return ([test.strip() for test in out.split('\n') if
             _is_valid_test_id(test)])


def _is_valid_test_id(test_str):
    """Return True if test_str looks like valid test id."""
    stripped = test_str.strip()
    return bool(stripped) and ' ' not in stripped


def add_tests_ids_for_filters(config_stack, args):
    """
    Add tests from Practitest filters to the config stack.

    :param config_stack: Test configuration data.
    :param args: Command line arguments
    :return: True if filters are specified and tests were successfully found,
             or no filters were specified. False when filters are specified and
             no matching tests are found.
    """
    if args.filters:
        pt_list = _get_test_ids_for_filters(config_stack, args.filters)
        args.suites.extend(pt_list)
        print('Filtered {n} tests from '
              'Practitest:\n{t}'.format(n=len(pt_list), t='\n'.join(pt_list)))
        return bool(pt_list)
    return True


def _get_test_ids_for_filters(config_stack, filters):
    """
    Return a list of tests for the specified filters parameters.

    :param config_stack: Test configuration data.
    :param filters: The command line filters parameter value.
    :return: List of test ids for the specified filters.
    """
    pt = PractitestSession(config_stack.get('_practitest_api_token'),
                           config_stack.get('_practitest_project_id'))
    tests = []
    for filter in filters.split(','):
        filter_type, filter_id = _validate_filter(filter)
        if filter_type == FILTER_ID_FILTER:
            result = pt.get_tests(filter_id)
        else:
            response = pt.get_test(filter_id)
            result = [response] if response else []
        ids = _get_ids_for_tests(result)
        if ids:
            tests.extend(ids)
        else:
            raise RuntimeError(
                '{f} returned no external ids.'.format(f=filter))
    return tests


def _get_ids_for_tests(tests):
    """Return list of External IDs from specified list of tests."""
    ids = []
    field_external_id = _get_external_id_field_id(tests)
    for test in tests:
        value = test
        for field in (field_external_id, FIELD_VALUE):
            value = value.get(field) if value else None
        if value:
            ids.append(value)
    return sorted(ids)


def _get_external_id_field_id(tests):
    """Return the field id of the External ID field in tests."""
    for test in tests:
        for key in test.keys():
            try:
                value = test[key][FIELD_NAME]
            except (KeyError, TypeError):
                continue
            else:
                if value == FIELD_EXTERNAL_ID:
                    return key
    raise RuntimeError('Could not find External ID field in response.')


def _validate_filter(filter):
    """
    Validate the specified filter, returning the type and id of the filter.

    Also check if an alias has been used and replace with the matching
    filter id.

    :param filter: Name of the filter to validate.
    :return: Tuple of filter type and id.
    :raises: ValueError if filter is not of valid format.

    """
    filter_id = FILTER_ALIAS.get(filter, filter)
    if ':' in filter_id:
        filter_type, filter_id = filter_id.split(':')
        if filter_type not in [FILTER_ID_FILTER, FILTER_ID_TEST]:
            raise ValueError(
                '{} is not a valid filter type'.format(filter_type))
    else:
        # No type is specified so assume that this is a filter id
        filter_type = FILTER_ID_FILTER
    if not filter_id.isdigit():
        raise ValueError('{} is not a valid filter id'.format(filter_id))
    return filter_type, filter_id


def _prepare_config(args):
    """Get config object for specified arguments."""
    config_stack = _get_config_stack(
        args.config, args.config_section, vars(args).get('silent', False))
    return config_stack


def get_camera_warning():
    return '- Please ensure device is raised from surface to allow rear ' \
           'camera to take a photograph during camera tests.'


def print_warnings():
    print_separator('*')
    logger.warning(get_camera_warning())
    print_separator('*')


def print_separator(char):
    logger.warning(char * 125)


def clean_output_directory(config_stack):
    """
    Clean the output directory specified in config_stack.

    This is required so that adt-run does not fail due to existing content
    in the output directory from a previous test run.

    """
    path = os.path.abspath(config_stack.get('output_dir'))
    if os.path.isdir(path):
        clean_dir(path)


def _get_config_stack(conf_name, config_section, silent=False):
    """Get the configuration values for the system tests.

    This function registers the declared options, opens the config store
    using the specified file name, adds the options coming from the command
    line, and asks the user for the missing options.

    :param conf_name: File name of config file to use.
    :param config_section: Section of config to use.
    :param silent: If True then don't prompt the user for any missing config
                   options, use default values instead. If False then prompt
                   the user for any missing option values.

    """
    config_stack = config.UbuntuSystemTestsConfig(
        conf_name, section=config_section)
    config_stack.get_config_from_file()
    config_stack.get_missing_options(silent)
    return config_stack


def run_setup(args):
    """Run setup commands to prepare the device for test execution.

    This will run the setup command on the device. If this causes a failure
    on the device, or a reboot is required, the device will reboot. The setup
    command will then be executed on the device again. If the second attempt
    at setup fails then the failure status will be reported and setup will
    stop.

    :param setup_cmd: Command to execute to set the device up.
    :param config_stack: Device configuration data.
    :return: Setup status code, 0 indicating setup was successful. Any error
             status encountered during setup will be translated to error
             status 16 from autopkgtest which indicates a testbed failure.
    """
    config_stack = _prepare_config(args)
    clean_output_directory(config_stack)
    add_tests_ids_for_filters(config_stack, args)
    target = targets.get_target(config_stack, args)
    target.connect()
    try:
        status = _run_setup(config_stack, args, target)
    finally:
        target.close()
    return status


def _run_setup(config_stack, args, target):
    """Run setup actions on target, rebooting and repeating if failure occurs
    on first attempt. Return status result.
    """
    series = target.series()
    ppa.valiate_ppa_args(args, series)
    count = 0
    status = 1
    while (status and count < 2):
        clean_output_directory(config_stack)
        status = target.run_setup().status
        if status and count < 1:
            # Error was reported after first setup attempt,
            # so reboot device and try again.
            print(
                'Rebooting device after setup completed with '
                'status {s}.'.format(s=status))
            target.reboot()
        count += 1
    return _run_post_setup(config_stack, args, target, status)


def _run_post_setup(config_stack, args, target, status):
    """Run post setup actions on the target and return status result."""
    if status:
        print('Setup failed with status: {}.'.format(status))
    elif args.cache_deps:
        # Download dependencies using autopkgtest
        adt_run_cmd = cmds.build_adt_run_test_command(
            config_stack, args, target)
        status = subprocess.call(adt_run_cmd)
    return status


def run_autopkg_test(config_stack, args):
    """ Run the autopkgtest in the device for the tests selected
    :param config_stack: The config objects
    :param args: command line arguments
    :return: the autopkgtest execution exit code
    """
    target = targets.get_target(config_stack, args)
    target.connect()
    adt_run_cmd = cmds.build_adt_run_test_command(config_stack, args, target)
    try:
        status = subprocess.call(adt_run_cmd)
    finally:
        target.close()
    return status


def generate_user_report(config_stack):
    """Create a human readable report from the subunit results."""
    results_dir = config_stack.get('output_dir')
    results.merge_subunit_results(results_dir)
    results.convert_subunit_results_to_pyunit(results_dir)
    results.print_results(results_dir)


if __name__ == '__main__':
    sys.exit(main())
