#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014-2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import subprocess

from ubuntu_system_tests.common import (
    config,
    ssh,
)
from ubuntu_system_tests.host import (
    debian,
    results,
)

_ADT_RUN_SHORT_TIMEOUT = 600  # 10 minutes.
_ADT_RUN_TEST_TIMEOUT = 86400  # 24 hours

TESTS_SETUP = ['setup']
TESTS_SYSTEMTESTS = ['systemtests']
# This is a dummy test id used to get default list of dependencies that
# excludes the large chromedriver dependencies used for webapps.
TESTS_UST_DEFAULT_DEPS = ['ubuntu_system_tests.tests.base']

STATUS_NO_CONNECTION = 1

legacy_cmds = None

XENIAL = '16.04'


def use_legacy_commands():
    """Return bool to indicate if autopkgtest legacy commands should be
    used."""
    global legacy_cmds
    if legacy_cmds is None:
        legacy_cmds = get_autopkgtest_version() < 4
    return legacy_cmds


def get_autopkgtest_version():
    """Return version of installed autopkgtest."""
    out = subprocess.check_output(
        ['dpkg-query', '--showformat=${Version}',
         '--show', 'autopkgtest'], timeout=30).decode().strip()
    if '.' in out:
        # Remove all but 1 decimal point
        dec_point = out.index('.')
        integer = _remove_alpha(out[:dec_point])
        remainder = _remove_alpha(out[dec_point+1:])
        ver = '{}.{}'.format(integer, remainder)
    else:
        ver = _remove_alpha(out)
    return float(ver)


def _remove_alpha(string):
    """Remove all alpha characters from string, leaving only digits."""
    return ''.join(char for char in string if char.isdigit())


def build_basic_adt_run_test_command(config_stack, args, verbose):
    """Build and return first part of adt-run command for test execution."""
    depends = debian.get_dependencies(args.suites or TESTS_UST_DEFAULT_DEPS)
    print('using test dependencies: {}'.format(' '.join(depends)))
    run_cmd = _get_autopkgtest_cmd_start(config_stack, verbose)
    if getattr(args, 'cache_deps', False):
        # This is a special case which only exists for setup.
        # In this case we run the debian/tests/setup script which is
        # used only to download required test dependencies.
        ctrl_file = debian.create_temp_control_file(
            tests=TESTS_SETUP, dependencies=depends)
    else:
        # This is standard test run which uses debian/tests/systemtests script.
        ctrl_file = debian.create_temp_control_file(
            tests=TESTS_SYSTEMTESTS, dependencies=depends)
    run_cmd.append('--override-control={}'.format(ctrl_file))
    run_cmd.extend([
        # Updating and installing can take a while.
        '--timeout-short={}'.format(_ADT_RUN_SHORT_TIMEOUT),
        # Regression tests can take up to 8 hours, no reason to cut short.
        '--timeout-test={}'.format(_ADT_RUN_TEST_TIMEOUT),
        '-B'])

    run_cmd.append(_get_src_tree_arg())

    run_cmd.extend([
        # Even if the config file on the host has a different name, on the
        # testbed it's copied with the default name.
        '--copy',
        '{src}:{dest}'.format(
            src=config_stack.file_path,
            dest=config.get_device_config_file_path()),
    ])
    return run_cmd


def build_adt_run_test_command(config_stack, args, target):
    """Build and return adt-run command for executing tests on device."""
    run_cmd = build_basic_adt_run_test_command(
        config_stack, args, args.verbose)
    run_cmd.extend(
        _get_network_setup_command(target, config_stack, args.network_ready))
    run_cmd.extend(
        _get_upgrade_setup_command(target, getattr(args, 'upgrade', False)))
    run_cmd.extend(_get_fs_ro_setup_command(target))
    run_cmd.extend(_get_set_tests_to_run_command(args.suites))
    run_cmd.extend(_get_set_config_section_command(args.config_section))
    touch_visualization = getattr(args, 'touch_visualization', False)
    run_cmd.extend(_get_set_touch_visualization(touch_visualization))
    run_cmd.extend(_get_apparmor_rules_update_setup_command())
    run_cmd.extend(ssh.get_ssh_config_commands())
    run_cmd.extend(_get_set_xdg_session_type_command())
    run_cmd.extend(_get_set_mir_socket_command())
    run_cmd.extend(_get_set_dbus_session_command(target, config_stack))

    return run_cmd + _get_virt_separator() + target.get_virt_command()


def _get_src_tree_arg():
    """Return command to specify src tree argument."""
    tree_arg = '--unbuilt-tree=' if use_legacy_commands() else ''
    return tree_arg + debian.get_debian_path()


def _get_virt_separator():
    """Return virt server command separator."""
    return ['---'] if use_legacy_commands() else ['--']


def _get_autopkgtest_cmd_start(config_stack, verbose):
    """Return the start of adt_run command."""
    executable = 'adt-run' if use_legacy_commands() else 'autopkgtest'
    output_dir = results.get_output_directory(config_stack)
    run_cmd = [executable, '--output-dir', output_dir]
    if verbose:
        run_cmd.append('-ddd')
    return run_cmd


def _get_upgrade_setup_command(target, upgrade):
    """Return setup command to upgrade csv list of components."""
    if upgrade:
        if target.snap_session():
            cmd = ''
            for name in upgrade.split(','):
                cmd += 'snap refresh --devmode --edge {}; '.format(name)
            cmd = cmd.strip()
        else:
            upgrade = upgrade.replace(',', ' ')
            cmd = (
                'apt-get -y --no-install-recommends install '
                '{}'.format(upgrade))
        return ['--setup-commands', _get_fs_rw_command(target) + cmd]
    return []


def _get_fs_ro_setup_command(target):
    """Return setup command to set file system read-only."""
    if target.config.get('mount_fs_ro', False):
        cmd = ('echo "Mounting file system ro"; '
               'sync; RETRY=0; until mount -o remount,ro / || '
               '[ $RETRY -gt 10 ]; '
               'do sleep 1; RETRY=$(($RETRY+1)); echo "mount attempt failed, '
               'trying again..."; done; if [ $RETRY -gt 10 ]; then exit '
               '{status}; fi; '.format(status='1'))
        return ['--setup-commands', cmd]
    # For cases where the device doesn't support a read-only file system,
    # use the ro-apt setup script to force apt into ro mode. This means that
    # autopkgtest will extract the deb files into /tmp rather than installing
    # packages onto the system.
    return ['--setup-commands', 'ro-apt']


def _get_fs_rw_command(target):
    """Return command to set file system writable."""
    if target.config.get('mount_fs_rw', False):
        return ('echo "Mounting file system rw"; '
                'sync; RETRY=0; until mount -o remount,rw / || '
                '[ $RETRY -gt 10 ]; '
                'do sleep 1; RETRY=$(($RETRY+1)); echo "mount attempt failed, '
                'trying again..."; done; if [ $RETRY -gt 10 ]; then exit '
                '{status}; fi; '.format(status='1'))
    return ''


def _get_network_setup_command(target, config_stack, network_ready):
    """ The network needs to be configured on the device
    This setup-command should be the first to run so that wi-fi is ready
    """
    if network_ready or not target.config.get('create_network', False):
        cmd = []
    else:
        cmd = [
            '--setup-commands',
            _get_device_network_command(
                config_stack.get('wifi_ssid'),
                config_stack.get('wifi_password')
            ),
        ]
    return cmd + _get_check_network_setup_command()


def _get_device_network_command(wifi_ssid, wifi_password):
    """Return command to configure wifi network on device.

    :param wifi_ssid: SSID of required wifi network
    :param wifi_password: Password of required wifi network

    """
    cmd_fmt = ('nmcli device wifi connect '
               '\'{ssid}\' password \'{password}\' 2>/dev/null || true')
    return cmd_fmt.format(ssid=wifi_ssid, password=wifi_password)


def _get_check_network_setup_command():
    return [
        '--setup-commands', 'echo \'Checking device network connection...\'; '
        'RETRY=0; until wget -q --spider launchpad.net || [ $RETRY -gt 60 ]; '
        'do sleep 1; RETRY=$(($RETRY+1)); echo "wget attempt $RETRY failed, '
        ' trying again..."; done; if [ $RETRY -gt 60 ]; then echo \'Device '
        'has no network connection!\'; exit {status}; fi'.format(
            status=STATUS_NO_CONNECTION)]


def _get_disable_auto_reboot_setup_command():
    """Return setup command to disable automatic rebooting by autopkgtest."""
    return [
        '--setup-commands',
        'if [ ! -f /run/autopkgtest_no_reboot.stamp ]; then '
        'touch /run/autopkgtest_no_reboot.stamp; fi; ']


def _get_apparmor_rules_update_setup_command():
    """Return setup command to update apparmor for autopilot introspection."""
    rules_file = '/var/cache/apparmor/click-ap.rules'
    click_hook = '/usr/bin/aa-clickhook'
    update_command = (
        'printf "dbus (receive, send) '
        'bus=session path=/com/canonical/Autopilot/**,\n'
        '/tmp/adt-run.** r,\n" > {rules}; {click} '
        '--include={rules}'.format(rules=rules_file, click=click_hook))
    return [
        '--setup-commands',
        ('if [ -e {ch} ] && [ ! -e {rf} ]; then {cmd}; '
         'fi'.format(ch=click_hook, rf=rules_file, cmd=update_command))
    ]


def _get_set_tests_to_run_command(tests_to_run):
    """Return command to set list of tests as environment variable."""
    tests = ' '.join(tests_to_run or [])
    return _get_set_env_var_adt_command('TESTS_TO_RUN', tests)


def _get_set_config_section_command(config_section):
    """Return command to set config section as environment variable."""
    return _get_set_env_var_adt_command('CONFIG_SECTION', config_section)


def _get_set_touch_visualization(touch_visualization):
    """Return command to set touch visualization as environment variable"""
    return _get_set_env_var_adt_command(
        'TOUCH_VISUALIZATION', str(touch_visualization).lower())


def _get_set_xdg_session_type_command():
    """Return command to set mir session type as environment variable."""
    return _get_set_env_var_adt_command('XDG_SESSION_TYPE', 'mir')


def _get_set_mir_socket_command():
    """Return command to set mir socket as environment variable."""
    return _get_set_env_var_adt_command(
        'MIR_SERVER_HOST_SOCKET', '/run/mir_socket')


def _get_set_env_var_adt_command(name, value):
    """Return command to set environment variable on device."""
    return ['--env={n}={v}'.format(n=name, v=value)]


def _get_set_dbus_session_command(target, config_stack):
    """Return command to set dbus session address environment variable."""
    if target.release() < XENIAL:
        return []
    user_id = target.user_id(config_stack.get('device_username'))
    return _get_set_env_var_adt_command(
        'DBUS_SESSION_BUS_ADDRESS',
        'unix:path=/run/user/{}/bus'.format(user_id))
