#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import requests

from softwareproperties.ppa import PPAException
from softwareproperties.shortcuts import ShortcutException
from softwareproperties.SoftwareProperties import shortcut_handler
from urllib.error import HTTPError

DEFAULT_PPA_LIST = ['ppa:canonical-platform-qa/selenium',
                    'ppa:canonical-platform-qa/ubuntu-system-tests']

PPA_ID = 'ppa:'
PPA_ID_SILO = 'silo:'
PPA_CI_TRAIN_SILO = 'ppa:ci-train-ppa-service/{id}'


def valiate_ppa_args(args, target_series):
    if getattr(args, 'ppa', False):
        ppa_ids = get_ppa_ids(args.ppa)
        validate_ppa_series(ppa_ids, target_series)


def get_ppa_ids(csv_list):
    """Convert any ppa in csv_list from 'silo:' format to 'ppa:' url format.
    Alternatively if no 'ppa:' prefix is used, then apply this by default.
    :param csv_list: Command line ppa argument specifying string comma
                     separated value list of additional ppas in either 'ppa:',
                    'silo:' format or with no prefix.
    :return: List of converted ppa urls in 'ppa:' format.
    """
    formatted_list = []
    if not csv_list:
        return formatted_list
    for ppa in csv_list.split(','):
        formatted_list.append(_get_formatted_ppa(ppa))
    return formatted_list


def validate_ppa_series(ppa_ids, target_series):
    """Validate each ppa in list is valid and return list of Ppa objects."""
    for ppa_id in ppa_ids:
        ppa_info = _get_ppa_info(ppa_id)
        if not ppa_info:
            raise NameError('{p} is not a valid ppa.'.format(p=ppa_id))
        ppa_dist_url = _get_ppa_dists_url(ppa_info)
        series_url = os.path.join(ppa_dist_url, target_series)
        response = requests.get(series_url)
        if not response.ok:
            raise ValueError(
                '{} does not publish '
                'for target series: {}'.format(ppa_id, target_series))


def _get_formatted_ppa(ppa):
    """Return correctly formatted ppa id starting with 'ppa:' prefix. This will
    convert 'silo:' or no prefix to the standard 'ppa:' prefix.
    :param ppa: Unformatted ppa string.
    :return: ppa id starting with 'ppa:' prefix.
    """
    if ppa.startswith(PPA_ID_SILO):
        return PPA_CI_TRAIN_SILO.format(id=ppa.split(':')[1])
    elif not ppa.startswith(PPA_ID):
        return PPA_ID + ppa
    return ppa


def _get_ppa_info(ppa):
    """Return ppa info dictionary, or None if not found.
    :param ppa: required ppa id.
    """
    try:
        shortcut = shortcut_handler(ppa)
    except (HTTPError, PPAException, ShortcutException):
        return None
    else:
        return shortcut.info()


def _get_ppa_dists_url(ppa_info):
    """Return url of ppa which shows all available series."""
    private = 's' if ppa_info.get('private') else ''
    owner, dist, name = ppa_info['reference'].split('/')
    if owner.startswith('~'):
        owner = owner[1:]
    url = 'http{p}://ppa.launchpad.net/{o}/{n}/{d}/dists/'
    return url.format(p=private, o=owner, n=name, d=dist)
