#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016-2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""This script is copied to target and executed to perform setup actions."""

import dbus
import json
import os
import pwd
import shutil
import subprocess
import sys
import time

# apt-cacher-ng definitions
APT_CACHER = 'apt-cacher-ng'
APT_CACHER_CACHE_DIR = '/var/cache/apt-cacher-ng'
APT_CACHER_LOG_DIR = '/var/log/apt-cacher-ng/'
APT_CACHER_UPSTART_INIT_FILE = '/etc/init/ust-apt-cacher-ng.conf'
APT_CACHER_SYSTEMD_INIT_FILE = '/etc/systemd/system/ust-apt-cacher-ng.service'
APT_CACHER_START_SCRIPT = '/usr/share/apt-cacher-ng/apt-cacher-ng-start.sh'
APT_PROXY = 'Acquire::http { Proxy "http://127.0.0.1:3142"; };'
APT_PROXY_FILE = '/etc/apt/apt.conf.d/01proxy'
DEFAULT_PERM = 0o755

QTTESTABILITY_AUTOPILOT = 'qttestability-autopilot'
INSTALL_LIST = [QTTESTABILITY_AUTOPILOT]

CITRAIN_OVERLAY_PPA = 'ppa:ci-train-ppa-service/stable-phone-overlay'
XENIAL = 16.04
YAKKETY = 16.10


class SetupRunner:

    """Class to run setup actions directly on the target device."""

    def __init__(self, config):
        self.config = config
        self.series = None
        self.release = None
        self._set_cache_dir()

    def _set_cache_dir(self):
        cache_dir = self.config.get('apt_cache_dir')
        if cache_dir:
            self.ext_cache_dir = os.path.join(cache_dir, 'cache')
            self.ext_logs_dir = os.path.join(cache_dir, 'logs')
        else:
            self.ext_cache_dir = None
            self.ext_logs_dir = None

    def run_setup_commands(self):
        self.setup_network()
        self.wait_for_network()
        self.add_ppas()
        self.update_apt()
        self.dist_upgrade()
        self.setup_unity8_session()
        self.install_test_packages()
        self.install_apt_cacher()
        with open("/etc/environment", "r") as environment:
            if "QT_LOAD_TESTABILITY" not in environment.read():
                self.enable_testability()

    def get_release(self):
        """Return float release number of running system."""
        if not self.release:
            self.release = float(subprocess.check_output(
                ['lsb_release', '-rs']).decode().strip())
        return self.release

    def get_series(self):
        """Return string release series name of runnning system."""
        if not self.series:
            self.series = subprocess.check_output(
                ['lsb_release', '-cs']).decode().strip()
        return self.series

    def update_apt(self):
        """Return command to apt-get update."""
        if self.config['update_apt']:
            self.run_update_apt()

    def run_update_apt(self):
        self.mount_fs_rw()
        subprocess.check_call(['apt-get', 'update'])

    def enable_testability(self):
        self.mount_fs_rw()
        with open("/etc/environment", "a") as outfile:
            outfile.write("QT_LOAD_TESTABILITY=1")

    def setup_network(self):
        """Return command to configure wifi network on device.

        :param wifi_ssid: SSID of required wifi network
        :param wifi_password: Password of required wifi network

        """
        if (self.config.get('create_network') and
           not self.config.get('network_ready')):
            subprocess.check_call(
                ['nmcli', 'device', 'wifi',
                 'connect', self.config['wifi_ssid'],
                 'password', self.config['wifi_password']])

    def wait_for_network(self):
        """Wait until network connection is active."""
        max_count = 60
        for count in range(max_count):
            status = subprocess.call(
                ['wget', '-q', '--spider', 'launchpad.net'])
            if status == 0:
                break
            elif count >= max_count - 1:
                raise ConnectionError('No network connection available')
            else:
                print('wget attempt {} failed, '
                      'trying again...'.format(count + 1))
                time.sleep(1)

    def mount_fs_ro(self):
        if self.config.get('mount_fs_ro') and self.is_fs_rw():
            self.mount_fs('ro')

    def mount_fs_rw(self):
        if self.config.get('mount_fs_rw') and self.is_fs_ro():
            self.mount_fs('rw')

    def mount_fs(self, mode):
        remount_param = 'remount,{}'.format(mode)
        for count in range(10):
            status = subprocess.call(
                ['mount', '-o', remount_param, '/'])
            if status:
                print('mount attempt failed, trying again...')
            else:
                return
        raise RuntimeError('Cannot mount file system {}.'.format(mode))

    def is_fs_ro(self):
        return self.is_fs('ro')

    def is_fs_rw(self):
        return self.is_fs('rw')

    def is_fs(self, mode):
        with open('/proc/mounts') as f:
            mounts = f.readlines()
        for mount in mounts:
            if ' / ext' in mount and mode in mount:
                return True
        return False

    def install_packages(self, packages):
        if packages and not self.are_packages_installed(packages):
            self.mount_fs_rw()
            subprocess.check_call(
                ['apt-get', '-y', '--no-install-recommends', 'install'] +
                packages)

    def install_snaps(self, snaps):
        if snaps:
            self.mount_fs_rw()
            for snap in snaps:
                cmd = 'refresh' if self.is_snap_installed(snap) else 'install'
                subprocess.check_call(
                    ['snap', cmd, '--devmode', '--edge', snap])

    def is_snap_installed(self, snap):
        return os.path.isdir(os.path.join('/snap', snap))

    def are_packages_installed(self, packages):
        output = subprocess.check_output(['apt-cache', 'policy'] + packages)
        return len(output) > 0 and 'Installed: (none)' not in output.decode()

    def install_test_packages(self):
        pkgs = INSTALL_LIST
        if self.config.get('install_tests', False):
            pkgs += self.config.get('test_dependencies', [])
        self.install_packages(pkgs)

    def is_dist_upgrade_required(self):
        user_acc = self._get_user_account_name()
        output = subprocess.check_output(
            ['su', '-', user_acc, '-c', 'apt-get dist-upgrade -s']).decode()
        return any(match in output for match in ['Inst', 'Conf', 'Remv'])

    def dist_upgrade(self):
        if self.config.get('dist_upgrade') and self.is_dist_upgrade_required():
            self.mount_fs_rw()
            self._dist_upgrade()

    def setup_unity8_session(self):
        """Install unity8 as the default session if required."""
        if self.get_release() < XENIAL:
            # unity8-desktop-session in only supported for xenial and later.
            # In this case we assume running vivid based ubuntu-touch which
            # already has unity8 running, so just return.
            return
        self.setup_snapd_https_proxy()
        if self.config.get('unity8_mode') == 'snap':
            # install unity8-session snap and set as default session
            self.install_packages(['unity8-session-snap'])
            subprocess.check_call(['unity8-snap-install'])
            session_name = 'unity8-snap'
        else:
            # install deb based unity8-desktop-session and set as default.
            # also install ubuntu-terminal-app which is tested as deb
            self.install_packages([
                'unity8-desktop-session',
                'ubuntu-terminal-app',
            ])
            # Also install applications as snaps, which can also be launched
            # in a deb based session
            self.install_snaps([
                'address-book-app',
                'camera-app',
                'gallery-app',
                'ubuntu-calculator-app',
                'ubuntu-calendar-app',
                'ubuntu-clock-app',
                'ubuntu-filemanager-app',
                'webbrowser-app',
            ])
            session_name = 'unity8'
        self.set_unity8_default_session(session_name)

    def setup_snapd_https_proxy(self):
        """Setup https proxy for snapd if its defined in config."""
        proxy = self.config.get('https_proxy')
        if proxy:
            conf = '/etc/systemd/system/snapd.service.d/proxy.conf'
            self._create_folder(os.path.dirname(conf))
            self._create_file(
                conf,
                '[Service]\n'
                'Environment=https_proxy={}\n'.format(proxy))
            subprocess.check_call(['systemctl', 'daemon-reload'])
            subprocess.check_call(['systemctl', 'restart', 'snapd.service'])

    def set_unity8_default_session(self, name):
        """Set unity8 to be the default session for user login."""
        self._create_file(
            '/etc/lightdm/lightdm.conf',
            '[Seat:*]\n'
            'user-session={name}\n'.format(name=name)
        )
        # Enable auto login for the user
        self.set_auto_login()

    def set_auto_login(self):
        """Set auto login to True for primary user."""
        user_id = self._get_user_account_id()
        bus = dbus.SystemBus()
        user_obj = bus.get_object(
            'org.freedesktop.Accounts',
            '/org/freedesktop/Accounts/User{}'.format(user_id))
        user_if = dbus.Interface(user_obj, 'org.freedesktop.Accounts.User')
        user_if.SetAutomaticLogin(True)

    def _get_user_account(self):
        for account in sorted(pwd.getpwall(), key=lambda x: x.pw_uid):
            if account.pw_uid > 500 and account.pw_dir.startswith('/home/'):
                return account

    def _get_user_account_name(self):
        return self._get_user_account().pw_name

    def _get_user_account_id(self):
        return self._get_user_account().pw_uid

    def _dist_upgrade(self):
        ppas = self.config.get('ppa_list')
        hold_list = self.config.get('dist_upgrade_hold_list')
        self._pre_dist_upgrade(ppas, hold_list)
        subprocess.check_call(
            ['apt-get', 'dist-upgrade', '--yes', '--force-yes'])
        subprocess.check_call(
            ['apt-get', 'autoremove', '--yes', '--force-yes'])
        self._post_dist_upgrade(ppas, hold_list)

    def _pre_dist_upgrade(self, ppas, hold_list):
        if ppas:
            self._add_ppa_config(ppas)
        if hold_list:
            subprocess.check_call(['apt-mark', 'hold'] + hold_list)

    def _post_dist_upgrade(self, ppas, hold_list):
        if hold_list:
            subprocess.check_call(['apt-mark', 'unhold'] + hold_list)
        if ppas:
            self._remove_ppa_config()

    def _add_ppa_config(self, ppas):
        policy = '/usr/sbin/policy-rc.d'
        tmp_policy = '/tmp/policy-rc.d'
        if os.path.isfile(policy):
            shutil.move(policy, tmp_policy)
        with open(policy, 'w') as f:
            f.write('exit 101')
        os.chmod(policy, 0o755)
        self._write_extra_ppas_pref_content(ppas)

    def _write_extra_ppas_pref_content(self, ppas):
        """Return command to create apt preferences file for upgrade.
        :param ppas: List of PPAs that have been added.
        :return: Command to create apt preferences file.
        """
        content = ''
        for ppa_obj in ppas:
            path = ppa_obj.split('/')[1]
            content += (
                'Package: *\n'
                'Pin: release o=*{p}*\n'
                'Pin-Priority:1100\n\n'.format(p=path)
            )
        content += ('Package: *\n'
                    'Pin: release a={s}*\n'
                    'Pin-Priority: 100\n').format(s=self.get_series())
        with open('/etc/apt/preferences.d/extra-ppas.pref', 'w') as f:
            f.write(content)

    def _remove_ppa_config(ppas):
        if os.path.isfile('/tmp/policy-rc.d'):
            shutil.move('/tmp/policy-rc.d', '/usr/sbin/policy-rc.d')
        else:
            os.remove('/usr/sbin/policy-rc.d')

    def add_ppas(self):
        for ppa in self.get_ppa_list():
            if not self._is_ppa_installed(ppa):
                self.mount_fs_rw()
                self._add_ppa(ppa)

    def get_ppa_list(self):
        """Return list of required PPAs to add. Include CI Train stable overlay
        PPA if release is younger than Y. Also include additional PPAs
        specified in config.
        """
        ppa = [CITRAIN_OVERLAY_PPA] if self.get_release() < YAKKETY else []
        return ppa + self.config.get('ppa_list')

    def _is_ppa_installed(self, ppa):
        ppa_id = ppa.split(':')[1]
        for root, dirs, files in os.walk('/etc/apt/sources.list.d'):
            for file in files:
                path = os.path.join(root, file)
                if self._is_ppa_match(path, ppa_id):
                    return True
        return False

    def _is_ppa_match(self, file_path, ppa_id):
        """Return True if apt source file contains required ppa id."""
        with open(file_path) as f:
            lines = f.readlines()
        for line in lines:
            if line.startswith('deb ') and ppa_id in line:
                return True
        return False

    def _add_ppa(self, ppa_id):
        subprocess.check_call(['apt-add-repository', '-y', ppa_id])

    def install_apt_cacher(self):
        if (self.config.get('cache_deps') and
                not self.are_packages_installed([APT_CACHER])):
            self.mount_fs_rw()
            self.install_packages([APT_CACHER])
            self._set_ext_cache_dir()
            self._create_file(APT_PROXY_FILE, APT_PROXY)
            self._apt_cacher_autostart()

    def _start_apt_cacher(self):
        cmds = self._get_apt_cacher_start_command().split(' ')
        subprocess.check_call(cmds)

    def _stop_apt_cacher(self):
        subprocess.call(['pkill', 'apt-cacher-ng'])

    def _set_ext_cache_dir(self):
        if self.config.get('apt_cache_dir'):
            self._stop_apt_cacher()
            self._create_folder(self.ext_logs_dir, APT_CACHER, DEFAULT_PERM)
            self._create_folder(self.ext_cache_dir, APT_CACHER, DEFAULT_PERM)
            self._create_folder(APT_CACHER_CACHE_DIR, APT_CACHER, DEFAULT_PERM)
            self._mount_dir(self.ext_logs_dir, APT_CACHER_LOG_DIR)
            self._mount_dir(self.ext_cache_dir, APT_CACHER_CACHE_DIR)
            self._start_apt_cacher()

    def _get_apt_cacher_start_command(self):
        """Return command to start apt-cacher-ng process."""
        return ('/usr/sbin/apt-cacher-ng -c /etc/apt-cacher-ng '
                'pidfile=/var/run/apt-cacher-ng/pid '
                'SocketPath=/var/run/apt-cacher-ng/socket '
                'foreground=0')

    def _apt_cacher_autostart(self):
        """Write the apt-cacher-ng conf file into /etc/init/apt-cacher-ng.conf
        to re-launch apt-cacher-ng on each subsequent reboot.

        This method assumes that upstart is used as the init system. If this
        changes in future then this method will need updating to accomodate
        that.
        """
        start_cmds = self._get_apt_cacher_autostart_commands()
        if start_cmds:
            # Create a startup script and register it to start on boot
            self._create_apt_cacher_autostart_script(start_cmds)
            if self._is_init_systemd():
                # For systemd based systems, apt-cacher-ng will start
                # automatically but there still may be file mount commands to
                # execute to change the storage directory for cached deps.
                # This will be defined in the target.json file.
                self._apt_cacher_systemd_autostart()
            else:
                # For upstart based systems, apt-cacher-ng does not start
                # automatically so we need to create a new job to do this
                # and perform any file mounts neccessary to change the
                # storage directory for cached deps.
                self._apt_cacher_upstart_autostart()

    def _get_init_system(self):
        return subprocess.check_output(
            ['ps', '-p1', '-o', 'comm', '--no-headers']).decode().strip()

    def _is_init_systemd(self):
        return self._get_init_system() == 'systemd'

    def _is_init_upstart(self):
        return self._get_init_system() == 'init'

    def _array_to_script(self, lines):
        """ Retrieve a string with \n characters based on a string list
        :param lines: A list with the lines
        :return: the string with the array elements and \n characters
        """
        return '\n'.join(lines)

    def _create_file(self, path, content=None, permission=DEFAULT_PERM):
        with open(path, 'w') as f:
            f.write(content)
        os.chmod(path, permission)

    def _create_folder(self, path, owner=None, permission=DEFAULT_PERM):
        if not os.path.isdir(path):
            os.makedirs(path)
        if owner:
            shutil.chown(path, user=owner)
        os.chmod(path, permission)

    def _mount_dir(self, target, source):
        cmds = self._get_mount_dir_command(target, source).split(' ')
        subprocess.check_call(cmds)

    def _get_mount_dir_command(self, target, source):
        """Return command to create a symlink between source and target
        directories."""
        return 'mount -o bind {target} {source}'.format(
            target=target, source=source)

    def _get_apt_cacher_autostart_commands(self):
        """Return list of commands required to start apt cacher."""
        cmds = self._get_apt_cacher_mount_commands()
        if self._is_init_upstart():
            # When installed, apt-cacher only autostarts on systemd by default.
            # So need to create a new job to launch it at boot for upstart.
            cmds.append(self._get_apt_cacher_start_command())
        return cmds

    def _create_apt_cacher_autostart_script(self, commands):
        """Create script to start apt-cacher including mounting cache and log
        directories if required.
        """
        script = self._array_to_script(commands)
        self._create_folder(os.path.dirname(APT_CACHER_START_SCRIPT))
        self._create_file(APT_CACHER_START_SCRIPT, script)

    def _get_apt_cacher_mount_commands(self):
        """Return commands to mount apt cacher cache and log directories."""
        cmds = []
        if self.ext_cache_dir:
            cmds.append(
                self._get_mount_dir_command(
                    self.ext_cache_dir, APT_CACHER_CACHE_DIR))
        if self.ext_logs_dir:
            cmds.append(
                self._get_mount_dir_command(
                    self.ext_logs_dir, APT_CACHER_LOG_DIR))
        return cmds

    def _apt_cacher_upstart_autostart(self):
        """Write the apt-cacher-ng conf file into apt-cacher-ng.conf
        to re-launch apt-cacher-ng on each subsequent reboot.

        This method is used to enable autostart on upstart based systems.
        """
        script = self._array_to_script([
            'start on runlevel [2345]',
            'exec {}'.format(APT_CACHER_START_SCRIPT)])
        self._create_file(APT_CACHER_UPSTART_INIT_FILE, script)

    def _apt_cacher_systemd_autostart(self):
        """Write the apt-cacher-ng conf file into apt-cacher-ng.service
        to re-launch apt-cacher-ng on each subsequent reboot.

        This method is used to enable autostart on systemd based systems.
        """
        script = self._array_to_script([
            '[Unit]',
            'Description=Autostart Apt Cacher NG',
            '[Service]',
            'Type=simple',
            'ExecStart=/bin/sh {}'.format(APT_CACHER_START_SCRIPT),
            '[Install]',
            'WantedBy=multi-user.target'])
        self._create_file(APT_CACHER_SYSTEMD_INIT_FILE, script)
        service_name = os.path.basename(APT_CACHER_SYSTEMD_INIT_FILE)
        subprocess.check_call(
            ['systemctl', 'enable', service_name])


def main(argv=None):
    args = argv or sys.argv[1:]
    config_path = args[0]
    with open(config_path, 'r') as f:
        config_json = json.loads(f.read())
    runner = SetupRunner(config_json)
    try:
        runner.run_setup_commands()
    finally:
        runner.mount_fs_ro()


if __name__ == '__main__':
    sys.exit(main())
