#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016-2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import errno
import glob
import json
import os
import pexpect
import shutil
import signal
import socket
import subprocess
import threading
import tempfile
import time

from abc import ABCMeta
from abc import abstractmethod
from fixtures import TempDir

from ubuntu_system_tests.host import (
    debian,
    ppa,
    qemu,
    qemu_utils,
)
from ubuntu_system_tests.host.executors import (
    Result,
    SSHExecutor,
    TIMEOUT_CMD,
    TIMEOUT_CONNECT,
)

TESTS_UST_DEFAULT = ['ubuntu_system_tests']

TARGET_ADB = 'adb'
TARGET_QEMU = 'qemu'
TARGET_SSH = 'ssh'
TARGETS = [TARGET_ADB, TARGET_QEMU, TARGET_SSH]

TIMEOUT_SETUP = 2400


def get_target(config_stack, args):
    """Return target object for required type. Either adb or qemu."""
    target_type = config_stack.get('target_type')
    if target_type not in TARGETS:
        raise ValueError('{} is not a valid target type.'.format(target_type))
    target_map = {
        TARGET_ADB: AdbTarget,
        TARGET_QEMU: QemuTarget,
        TARGET_SSH: SSHTarget,
    }
    return target_map[target_type](config_stack, args)


class Target:
    __metaclass__ = ABCMeta
    """Base class for all targets."""

    def __init__(self, config_stack, args):
        self.config_stack = config_stack
        self.args = args
        self.config = self.get_config(config_stack, args)

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def put(self, src, dst):
        pass

    @abstractmethod
    def get(self, src, dst):
        pass

    @abstractmethod
    def run(self, *args, **kwargs):
        pass

    @abstractmethod
    def run_sudo(self, *args, **kwargs):
        pass

    @abstractmethod
    def shutdown(self):
        pass

    @abstractmethod
    def reboot(self):
        pass

    def snap_session(self):
        """Return boolean indicating if unity8 is running from snap session."""
        return self.run(
            'RETRY=0; '
            'until pid=$(pidof unity8) || [ $RETRY -gt 30 ]; do '
            'sleep 1; RETRY=$(($RETRY+1)); '
            'echo "unity8 not running, waiting..."; '
            'done; '
            'if [ $RETRY -gt 30 ]; then exit 1; fi; '
            'readlink -f /proc/$pid/exe | '
            'grep -q "^/snap/"', log_stdout=False).status == 0

    def series(self):
        return self.run('lsb_release -cs', log_stdout=False).output.strip()

    def release(self):
        return self.run('lsb_release -rs', log_stdout=False).output.strip()

    def user_id(self, user):
        """Return user id for specified user name."""
        return self.run(
            'id -u {}'.format(user), log_stdout=False).output.strip()

    @abstractmethod
    def get_virt_command(self):
        pass

    def run_setup(self):
        try:
            result = self._run_target_setup()
        except TimeoutError:
            result = Result()
            result.status = 1
        return result

    def _push_setup_script(self):
        src = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), 'target_setup.py')
        dst = os.path.join('/tmp/target_setup.py')
        self.put(src, dst)
        if self.run_sudo('chmod 744 {}'.format(dst)).status:
            raise RuntimeError(
                'Could not set script permission. '
                'Check password "{}" is correct.'.format(self.password))

    def _push_setup_config(self):
        with TempDir() as d:
            src = os.path.join(d.path, 'target_config.json')
            with open(src, 'w') as f:
                f.write(json.dumps(self.config))
            self.put(src, '/tmp/target_config.json')

    def _run_target_setup(self):
        self._push_setup_script()
        self._push_setup_config()
        return self.run_sudo(
            '/usr/bin/python3 /tmp/target_setup.py /tmp/target_config.json',
            timeout=TIMEOUT_SETUP)

    def get_config(self, config_stack, args):
        """Select the required config items to pass to the target setup
        script.
        """
        # Read target config from json
        config = self.get_target_config()
        # Add selected options from command line args
        for key in ['cache_deps', 'dist_upgrade', 'install_tests',
                    'network_ready', 'update_apt', 'verbose']:
            config[key] = getattr(args, key, False)
        # Add options from config_stack
        for key in ['wifi_ssid', 'wifi_password', 'unity8_mode']:
            config[key] = config_stack.get(key)
        # Add list of required test dependencies
        config['test_dependencies'] = debian.get_dependencies(
            getattr(args, 'suites', TESTS_UST_DEFAULT), control_fmt=False)
        # Add list of ppas
        ppa_list = ppa.DEFAULT_PPA_LIST
        ppa_args = ppa.get_ppa_ids(args.ppa) if hasattr(args, 'ppa') else []
        ppa_list += ppa_args
        config['ppa_list'] = ppa_list
        if ppa_args:
            # Do a dist upgrade if there are extra ppas added from cmdline args
            config['dist_upgrade'] = True
        # Add https_proxy from environment if it exists
        config['https_proxy'] = os.environ.get('https_proxy', False)
        return config

    @staticmethod
    def get_target_json():
        path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), 'targets.json')
        with open(path, 'r') as f:
            return json.loads(f.read())

    def get_target_config(self):
        target_json = self.get_target_json()
        return target_json[self.name]


class SSHTarget(Target):

    name = TARGET_SSH
    port_from = 22220
    port_to = 23000

    def __init__(self, config_stack, args):
        super().__init__(config_stack, args)
        self.executor = SSHExecutor()
        self.username = config_stack.get('device_username')
        self.password = config_stack.get('device_password')
        self.connected = False
        self.key_file = None
        self.device_ip = getattr(args, 'ip', None) or config_stack.get(
            'device_ip', 'localhost')
        self.port = int(getattr(args, 'port', None) or config_stack.get(
            'device_port', '-1'))

    def connect(self, timeout=TIMEOUT_CONNECT):
        if not self.connected:
            self._enable_ssh()
            self.executor.connect(
                self.username, self.password, self.port, self.device_ip,
                timeout)
            self.connected = True

    def close(self):
        if self.connected:
            self.executor.close()
            self.connected = False

    def reboot(self):
        self.executor.reboot()

    def shutdown(self):
        self.executor.shutdown()

    def put(self, src, dst):
        self.executor.put(src, dst)

    def run(self, command, timeout=TIMEOUT_CMD, log_stdout=True):
        return self.executor.run(command, timeout, log_stdout)

    def run_sudo(self, command, timeout=TIMEOUT_CMD, log_stdout=True):
        return self.executor.run_sudo(command, timeout, log_stdout)

    def find_free_port(self):
        for port in range(self.port_from, self.port_to):
            try:
                s = socket.create_connection(('127.0.0.1', port))
            except socket.error as e:
                if e.errno == errno.ECONNREFUSED:
                    # This means port is not currently used, so use this one
                    self.port = port
                    return
                else:
                    pass
            else:
                # port is already taken
                s.close()
        raise RuntimeError('Could not find free port for SSH connection.')

    def get_virt_command(self):
        debug = ['--debug'] if self.args.verbose else []
        return ['ssh', '--port', str(self.port),
                '--login', self.username,
                '--password', self.password,
                '--identity', self.key_file,
                '--hostname',
                self.device_ip] + debug

    def _enable_ssh(self):
        """Enable ssh using public key."""
        self._wait_for_device()
        self._get_ssh_id_path()
        if not self._try_public_key_login():
            self._update_device_host_key()
            self._copy_ssh_id_to_device()
            self._verify_ssh_connect()

    def _wait_for_device(self, timeout=TIMEOUT_CONNECT):
        end = time.time() + timeout
        while time.time() < end:
            try:
                s = socket.create_connection(
                    (self.device_ip, str(self.port)), timeout)
            except ConnectionRefusedError:
                time.sleep(1)
            else:
                s.close()
                return
        raise TimeoutError(
            'Could not connect to {} '
            'port {}.'.format(self.device_ip, self.port))

    def _update_device_host_key(self):
        hosts_path = os.path.expanduser('~/.ssh/known_hosts')
        subprocess.call([
            'ssh-keygen', '-f', hosts_path, '-R',
            '[{}]:{}'.format(self.device_ip, self.port)])

    def _copy_ssh_id_to_device(self):
        pub_path = '{}.pub'.format(self.key_file)
        home_ssh = '/home/{u}/.ssh'.format(u=self.username)
        authorized_keys = os.path.join(home_ssh, 'authorized_keys')
        self._run(['mkdir', '-p', home_ssh])
        self._put(pub_path, authorized_keys)
        self._run(
            ['chown', '{u}:{u}'.format(u=self.username), '-R', home_ssh])
        self._run(['chmod', '700', home_ssh])
        self._run(['chmod', '600', authorized_keys])

    def _get_ssh_id_path(self):
        match = False
        for id in ['~/.ssh/id_rsa', '~/.ssh/id_autopkgtest']:
            path = os.path.expanduser(id)
            if os.path.exists(path):
                match = True
                break
        if not match:
            subprocess.check_call([
                'ssh-keygen', '-q', '-t', 'rsa', '-f', path, '-N', ''])
        self.key_file = path

    def _try_public_key_login(self):
        """Try and log in using public key. If this succeeds then there is no
        need to do any further ssh setup.
        :return: True if login was successful, False otherwise
        """
        cmd = ' '.join(['ssh', '-p', str(self.port),
                        '-o', 'UserKnownHostsFile=/dev/null',
                        '-o', 'StrictHostKeyChecking=no',
                        '-i', self.key_file, '-l', self.username,
                        self.device_ip])
        child = pexpect.spawn(cmd)
        try:
            index = child.expect(
                ['\$', 'password', 'denied'], timeout=TIMEOUT_CONNECT)
        except (pexpect.exceptions.TIMEOUT, pexpect.exceptions.EOF):
            index = -1
        finally:
            child.close()
        return index == 0

    def _verify_ssh_connect(self):
        """Verify that an ssh connection can be established without using
        password.
        """
        for count in range(20):
            if self._try_public_key_login():
                return
            else:
                time.sleep(1)
        raise RuntimeError('Could not create ssh connection')

    def _run(self, commands, timeout=TIMEOUT_CMD):
        """Run a command setting up an ssh connection using password."""
        ssh_cmd = [
            'ssh', '-o', 'StrictHostKeyChecking=no', '-p', str(self.port),
            '{}@{}'.format(self.username, self.device_ip)]
        self._run_with_password(ssh_cmd + commands, self.password, timeout)

    def _put(self, src, dst, timeout=TIMEOUT_CMD):
        """Put file onto device setting up an ssh connection using password."""
        scp_cmd = [
            'scp', '-o', 'StrictHostKeyChecking=no', '-P', str(self.port),
            src, '{}@{}:{}'.format(self.username, self.device_ip, dst)]
        self._run_with_password(scp_cmd, self.password, timeout)

    def _run_with_password(self, commands, password, timeout=TIMEOUT_CMD):
        """Run command expecting a password prompt to be displayed."""
        command = ' '.join(commands)
        child = pexpect.spawn(command)
        try:
            child.expect('password', timeout=timeout)
        except pexpect.exceptions.EOF:
            # No password prompt is displayed, so just continue
            pass
        else:
            child.sendline(password)
            if child.expect([pexpect.EOF, 'denied'], timeout=timeout):
                raise PermissionError(
                    'Check password is correct: "{}"'.format(password))
        finally:
            child.close()
        if child.exitstatus:
            raise RuntimeError('Error running {}'.format(command))


class AdbTarget(SSHTarget):

    name = TARGET_ADB

    def __init__(self, config_stack, args):
        super().__init__(config_stack, args)
        self.serial = self._get_serial_argument()

    def connect(self):
        self._wait_for_adb_device()
        self.find_free_port()
        self._enable_port_forwarding()
        self._wait_for_device()
        super().connect(timeout=1)

    def close(self):
        super().close()
        self._run_adb_command(
            'forward', '--remove', 'tcp:{}'.format(self.port))

    def reboot(self):
        self.close()
        self._run_adb_command('shell', 'sync')
        self._run_adb_command('reboot')
        self.connect()

    def _run_adb_command(self, *commands):
        subprocess.check_call(
            ['adb'] + self.serial + list(commands), stdout=subprocess.DEVNULL)

    def _run(self, commands, timeout=30):
        """Run a command through adb shell"""
        self._run_adb_command('shell', ' '.join(commands))

    def _put(self, src, dst, timeout=30):
        """Put file onto device using adb push."""
        self._run_adb_command('push', src, dst)

    def _wait_for_adb_device(self):
        """Wait for required adb device to be online."""
        self._run_adb_command('wait-for-device')

    def _enable_port_forwarding(self):
        self._run_adb_command('forward', 'tcp:{}'.format(self.port), 'tcp:22')
        self._run_adb_command(
            'shell', 'gdbus', 'call', '-y', '-d',
            'com.canonical.PropertyService', '-o',
            '/com/canonical/PropertyService', '-m',
            'com.canonical.PropertyService.SetProperty', 'ssh', 'true')

    def _get_serial_argument(self):
        """Return serial commands from config or empty list if not specified"""
        device_serial = self.config_stack.get('device_serial')
        if device_serial:
            return ['-s', device_serial]
        else:
            # No serial in the config, run the commands as if only one device
            # is connected.
            return []

    def get_virt_command(self):
        """Return command parameters for ssh connection to requried device."""
        device_password = self.config_stack.get('device_password')
        return (['ssh', '-s', 'adb', '--', '-p', device_password] +
                self._get_serial_argument())


class QemuTarget(SSHTarget):

    name = TARGET_QEMU

    def __init__(self, config_stack, args):
        super().__init__(config_stack, args)
        self.qemu_runner = None
        self.work_dir = None
        self.overlay = None
        self._set_qemu_launch_opts()

    def _set_qemu_launch_opts(self):
        self.find_free_port()
        cache_data_path = self.config_stack.get('cache_data_path')
        img_path = self.args.img or os.path.join(cache_data_path, 'img')
        if os.path.isfile(img_path):
            # A file is specified so use that one
            self.img_file = img_path
        elif os.path.isdir(img_path):
            # Get the most recent img file from the directory
            self.img_file = max(
                glob.iglob(os.path.join(img_path, '*.img')),
                key=os.path.getctime)
        else:
            raise ValueError('{} does not exist.'.format(img_path))
        self.cpu = self.config_stack.get('qemu_cpu_count')
        self.ram = qemu_utils.get_ram_kb(
            self.config_stack.get('qemu_ram_alloc'))
        self.headless = self.args.headless or self.config_stack.get(
            'qemu_headless').lower() == 'true'
        if not getattr(self.args, 'no_overlay', True):
            # No-overlay is False, so we need to use overlay
            self.overlay = os.path.join(cache_data_path, 'tmp/overlay.img')

    def connect(self):
        self.qemu_runner = self.launch_qemu()
        super().connect()

    def close(self):
        try:
            self.shutdown()
        except PermissionError:
            print('Shutdown sudo command failed. '
                  'Check password: "{}".'.format(self.password))
            self.stop_qemu()
        finally:
            self.qemu_runner.join()
        shutil.rmtree(self.work_dir)
        self.qemu_runner = None
        self.work_dir = None
        super().close()

    def reboot(self):
        self.close()
        self.connect()

    def stop_qemu(self):
        pid_file = os.path.join(self.work_dir, 'qemu.pid')
        with open(pid_file) as f:
            pid = int(f.read().strip())
        os.kill(pid, signal.SIGTERM)

    def launch_qemu(self):
        """Boot the qemu from a different thread to stop this thread from being
        blocked whilst the qemu is running.
        """
        self.work_dir = tempfile.mkdtemp()
        runner = threading.Thread(
            target=self._launch_qemu, args=(self.work_dir,))
        runner.start()
        return runner

    def _launch_qemu(self, work_dir):
        qemu.launch_qemu(
            work_dir, self.img_file, self.ram, self.cpu, self.headless,
            port=self.port, overlay=self.overlay)

    def wait_for_stopped(self):
        """Wait until the qemu process is stopped."""
        self.qemu_runner.join()
