# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import testscenarios

from json.decoder import JSONDecodeError
from unittest.mock import (
    call,
    mock_open,
    patch,
)

from ubuntu_system_tests.host.target_setup import (
    main as target_setup_main,
    SetupRunner,
)
from ubuntu_system_tests.selftests.utils import ConfigBaseTestCase


class TestSetupRunnerNetwork(ConfigBaseTestCase,
                             testscenarios.TestWithScenarios):

    scenarios = [
        ('no create network', {
            'config': {'create_network': False,
                       'network_ready': False},
            'call_args': []
        }),
        ('no create network already setup', {
            'config': {'create_network': False,
                       'network_ready': True},
            'call_args': []
        }),
        ('network ready', {
            'config': {'create_network': True,
                       'network_ready': True},
            'call_args': []
        }),
        ('create network', {
            'config': {'create_network': True,
                       'network_ready': False,
                       'wifi_ssid': 'wifi-ssid-1',
                       'wifi_password': 'wifi-password-1'},
            'call_args': ['nmcli', 'device', 'wifi', 'connect', 'wifi-ssid-1',
                          'password', 'wifi-password-1']
        }),
    ]

    @patch('subprocess.check_call')
    def test_check_network_setup_from_config(self, mock_check_call):
        runner = SetupRunner(self.config)
        runner.setup_network()
        if self.call_args:
            mock_check_call.assert_called_once_with(self.call_args)
        else:
            mock_check_call.assert_not_called()


class TestSetupRunnerMountFsRo(ConfigBaseTestCase,
                               testscenarios.TestWithScenarios):

    scenarios = [
        ('mount ro', {
            'config': {'mount_fs_ro': True},
            'is_fs_rw': True,
            'call_args': ['mount', '-o', 'remount,ro', '/']
        }),
        ('already ro', {
            'config': {'mount_fs_ro': True},
            'is_fs_rw': False,
            'call_args': []
        }),
        ('no mount', {
            'config': {'mount_fs_ro': False},
            'is_fs_rw': True,
            'call_args': []
        }),
    ]

    @patch('subprocess.call', return_value=0)
    def test_mount_fs_ro(self, mock_call):
        runner = SetupRunner(self.config)
        with patch(
                'ubuntu_system_tests.host.target_setup.SetupRunner.is_fs_rw',
                return_value=self.is_fs_rw) as mock_is_fs_rw:
            runner.mount_fs_ro()
            if self.call_args:
                mock_call.assert_called_once_with(self.call_args)
            else:
                mock_call.assert_not_called()

            if self.config['mount_fs_ro']:
                mock_is_fs_rw.assert_called_once_with()
            else:
                mock_is_fs_rw.assert_not_called()


class TestSetupRunnerMountFsRw(ConfigBaseTestCase,
                               testscenarios.TestWithScenarios):

    scenarios = [
        ('mount rw', {
            'config': {'mount_fs_rw': True},
            'is_fs_ro': True,
            'call_args': ['mount', '-o', 'remount,rw', '/']
        }),
        ('already rw', {
            'config': {'mount_fs_rw': True},
            'is_fs_ro': False,
            'call_args': []
        }),
        ('no mount', {
            'config': {'mount_fs_rw': False},
            'is_fs_ro': True,
            'call_args': []
        }),
    ]

    @patch('subprocess.call', return_value=0)
    def test_mount_fs_rw(self, mock_call):
        runner = SetupRunner(self.config)
        with patch(
                'ubuntu_system_tests.host.target_setup.SetupRunner.is_fs_ro',
                return_value=self.is_fs_ro) as mock_is_fs_ro:
            runner.mount_fs_rw()
            if self.call_args:
                mock_call.assert_called_once_with(self.call_args)
            else:
                mock_call.assert_not_called()

            if self.config['mount_fs_rw']:
                mock_is_fs_ro.assert_called_once_with()
            else:
                mock_is_fs_ro.assert_not_called()


class TestSetupRunnerInstallPackages(ConfigBaseTestCase,
                                     testscenarios.TestWithScenarios):

    scenarios = [
        ('not already installed', {
            'to_install': ['pkg_1', 'pkg_2'],
            'already_installed': False,
            'call_args': ['apt-get', '-y', '--no-install-recommends',
                          'install', 'pkg_1', 'pkg_2']
        }),
        ('already installed', {
            'to_install': ['pkg_1', 'pkg_2'],
            'already_installed': True,
            'call_args': []
        }),
    ]

    @patch('subprocess.check_call', return_value=0)
    @patch('ubuntu_system_tests.host.target_setup.SetupRunner.mount_fs_rw')
    def test_install_packages(self, mock_mount_fs_rw, mock_check_call):
        runner = SetupRunner({})
        with patch(
                'ubuntu_system_tests.host.target_setup.SetupRunner.'
                'are_packages_installed',
                return_value=self.already_installed
                ) as mock_are_packages_installed:
            runner.install_packages(self.to_install)
            mock_are_packages_installed.assert_called_once_with(
                self.to_install)
            if self.already_installed:
                mock_check_call.assert_not_called()
                mock_mount_fs_rw.assert_not_called()
            else:
                mock_check_call.assert_called_once_with(self.call_args)
                mock_mount_fs_rw.assert_called_once_with()


class TestSetupRunnerDistUpgrade(ConfigBaseTestCase,
                                 testscenarios.TestWithScenarios):

    scenarios = [
        ('upgrade no ppas no hold', {
            'config': {'dist_upgrade': True,
                       'dist_upgrade_hold_list': [],
                       'ppa_list': []},
            'upgrade_required': True,
            'calls': [['apt-get', 'dist-upgrade', '--yes', '--force-yes'],
                      ['apt-get', 'autoremove', '--yes', '--force-yes']]
        }),
        ('no upgrade required', {
            'config': {'dist_upgrade': True,
                       'dist_upgrade_hold_list': [],
                       'ppa_list': []},
            'upgrade_required': False,
            'calls': []
        }),
        ('no upgrade configured', {
            'config': {'dist_upgrade': False,
                       'dist_upgrade_hold_list': [],
                       'ppa_list': []},
            'upgrade_required': True,
            'calls': []
        }),
        ('upgrade with hold', {
            'config': {'dist_upgrade': True,
                       'dist_upgrade_hold_list': ['hold_pkg_1'],
                       'ppa_list': []},
            'upgrade_required': True,
            'calls': [['apt-mark', 'hold', 'hold_pkg_1'],
                      ['apt-get', 'dist-upgrade', '--yes', '--force-yes'],
                      ['apt-get', 'autoremove', '--yes', '--force-yes'],
                      ['apt-mark', 'unhold', 'hold_pkg_1']]
        }),
        ('upgrade with ppa', {
            'config': {'dist_upgrade': True,
                       'dist_upgrade_hold_list': [],
                       'ppa_list': ['ppa:user/testppa']},
            'upgrade_required': True,
            'calls': [['apt-get', 'dist-upgrade', '--yes', '--force-yes'],
                      ['apt-get', 'autoremove', '--yes', '--force-yes']]
        }),
    ]

    @patch('subprocess.check_call', return_value=0)
    @patch('ubuntu_system_tests.host.target_setup.SetupRunner.mount_fs_rw')
    @patch('ubuntu_system_tests.host.target_setup.SetupRunner._add_ppa_config')
    @patch(
        'ubuntu_system_tests.host.target_setup.SetupRunner._remove_ppa_config')
    def test_dist_upgrade(self, mock_rm_ppa_config, mock_add_ppa_config,
                          mock_mount_fs_rw, mock_check_call):
        runner = SetupRunner(self.config)
        with patch(
                'ubuntu_system_tests.host.target_setup.SetupRunner.'
                'is_dist_upgrade_required',
                return_value=self.upgrade_required) as mock_upgrade_required:
            runner.dist_upgrade()
            if self.config['dist_upgrade']:
                mock_upgrade_required.assert_called_once_with()
            else:
                mock_upgrade_required.assert_not_called()
            expected_calls = [call(c) for c in self.calls]
            mock_check_call.assert_has_calls(expected_calls)
            if self.config['ppa_list']:
                mock_add_ppa_config.assert_called_once_with(
                    self.config['ppa_list'])
                mock_rm_ppa_config.assert_called_once_with()


class TestSetupRunnerAptCacher(ConfigBaseTestCase,
                               testscenarios.TestWithScenarios):

    scenarios = [
        ('install apt-cacher', {
            'config': {'cache_deps': True,
                       'mount_fs_rw': True,
                       'mount_fs_ro': True},
            'already_installed': False,
            'is_fs_rw': False
        }),
        ('no install apt-cacher', {
            'config': {'cache_deps': False},
            'already_installed': False,
            'is_fs_rw': False
        }),
        ('apt-cacher already installed', {
            'config': {'cache_deps': True,
                       'mount_fs_rw': True,
                       'mount_fs_ro': True},
            'already_installed': True,
            'is_fs_rw': False
        }),
        ('install fs already rw', {
            'config': {'cache_deps': True,
                       'mount_fs_rw': True,
                       'mount_fs_ro': True},
            'already_installed': False,
            'is_fs_rw': True
        }),
        ('install no fs rw', {
            'config': {'cache_deps': True,
                       'mount_fs_rw': False,
                       'mount_fs_ro': False},
            'already_installed': False,
            'is_fs_rw': False
        }),
    ]

    @patch('subprocess.call', return_value=0)
    @patch('ubuntu_system_tests.host.target_setup.SetupRunner.'
           'install_packages')
    @patch('ubuntu_system_tests.host.target_setup.SetupRunner.mount_fs_rw')
    @patch('ubuntu_system_tests.host.target_setup.SetupRunner.'
           '_set_ext_cache_dir')
    @patch('ubuntu_system_tests.host.target_setup.SetupRunner._create_file')
    @patch('ubuntu_system_tests.host.target_setup.SetupRunner.'
           '_apt_cacher_autostart')
    def test_install_apt_cacher(self, mock_autostart, mock_create_file,
                                mock_cache_dir, mock_mount_rw, mock_inst_pkgs,
                                mock_call):
        runner = SetupRunner(self.config)
        with patch(
                'ubuntu_system_tests.host.target_setup.SetupRunner.is_fs_rw',
                return_value=self.is_fs_rw):
            with patch('ubuntu_system_tests.host.target_setup.SetupRunner.'
                       'are_packages_installed',
                       return_value=self.already_installed) as mock_pkgs_inst:
                runner.install_apt_cacher()
                if self.config['cache_deps']:
                    mock_pkgs_inst.assert_called_once_with(['apt-cacher-ng'])
                    if self.already_installed:
                        mock_mount_rw.assert_not_called()
                        mock_inst_pkgs.assert_not_called()
                        mock_cache_dir.assert_not_called()
                        mock_create_file.assert_not_called()
                        mock_autostart.assert_not_called()
                    else:
                        mock_mount_rw.assert_called_once_with()
                        mock_inst_pkgs.assert_called_once_with(
                            ['apt-cacher-ng'])
                        mock_cache_dir.assert_called_once_with()
                        mock_create_file.assert_called_once_with(
                            '/etc/apt/apt.conf.d/01proxy',
                            'Acquire::http { Proxy "http://127.0.0.1:3142"; '
                            '};')
                        mock_autostart.assert_called_once_with()
                else:
                    mock_inst_pkgs.assert_not_called()
                    mock_mount_rw.assert_not_called()
                    mock_cache_dir.assert_not_called()
                    mock_create_file.assert_not_called()
                    mock_autostart.assert_not_called()


@patch('ubuntu_system_tests.host.target_setup.SetupRunner.run_setup_commands')
@patch('ubuntu_system_tests.host.target_setup.SetupRunner.mount_fs_ro')
class TestSetupRunnerMain(ConfigBaseTestCase):

    def test_setup_runner_main_valid_json(self, mock_mount_fs_ro,
                                          mock_run_setup):
        json = ('{"apt_cache_dir": "", '
                '"create_network": false, '
                '"mount_fs_ro": false, '
                '"mount_fs_rw": false, '
                '"dist_upgrade_hold_list": []}')
        with patch('builtins.open', mock_open(read_data=json)) as o:
            target_setup_main(['/tmp/test_config.json'])
            o.assert_called_once_with('/tmp/test_config.json', 'r')
            mock_run_setup.assert_called_once_with()
            mock_mount_fs_ro.assert_called_once_with()

    def test_setup_runner_main_invalid_json(self, mock_mount_fs_ro,
                                            mock_run_setup):
        json = ('"apt_cache_dir": "", '
                '"create_network": false, '
                '"mount_fs_ro": false, '
                '"mount_fs_rw": false, '
                '"dist_upgrade_hold_list": []')
        with patch('builtins.open', mock_open(read_data=json)):
            with self.assertRaises(JSONDecodeError):
                target_setup_main(['/tmp/test_config.json'])

    def test_mount_ro_called_after_failure(self, mock_mount_fs_ro,
                                           mock_run_setup):
        json = ('{"apt_cache_dir": "", '
                '"create_network": true, '
                '"mount_fs_ro": false, '
                '"mount_fs_rw": false, '
                '"dist_upgrade_hold_list": []}')
        mock_run_setup.side_effect = ConnectionError
        with patch('builtins.open', mock_open(read_data=json)) as o:
            with self.assertRaises(ConnectionError):
                target_setup_main(['/tmp/test_config.json'])
            o.assert_called_once_with('/tmp/test_config.json', 'r')
            mock_run_setup.assert_called_once_with()
            mock_mount_fs_ro.assert_called_once_with()
