# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from errno import ECONNREFUSED
from socket import error as socket_error
from unittest import mock

from ubuntu_system_tests.common import config
from ubuntu_system_tests.host import targets
from ubuntu_system_tests.selftests.utils import (
    ConfigBaseTestCase,
    DummyArgs,
)


class TestTargets(ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()

    def test_load_target_json(self):
        self.assertTrue(bool(targets.Target.get_target_json()))

    @mock.patch('ubuntu_system_tests.host.debian.get_dependencies',
                return_value=['dep1'])
    def test_get_target_adb(self, mock_get_deps):
        self.set_config_values(
            device_serial='abc123', device_password='1234',
            device_username='test_user', target_type=targets.TARGET_ADB)
        args = DummyArgs(port='2222')
        target = targets.get_target(self.config_stack, args)
        self.assertEqual(target.name, targets.TARGET_ADB)
        self.assertEqual(target.serial, ['-s', 'abc123'])
        self.assertEqual(target.username, 'test_user')
        self.assertEqual(target.password, '1234')

    @mock.patch('os.path.isfile', return_value=True)
    @mock.patch('ubuntu_system_tests.host.debian.get_dependencies',
                return_value=['dep1'])
    def test_get_target_qemu(self, mock_get_deps, mock_isfile):
        self.set_config_values(
            qemu_cpu_count='3', qemu_ram_alloc='4096',
            device_username='test_qemu_user', device_password='56789',
            cache_data_path='/tmp/cache/data', target_type=targets.TARGET_QEMU)
        args = DummyArgs(img='/tmp/disk.img', headless=False,
                         overlay='overlay.img', port='-1')
        excp = socket_error()
        excp.errno = ECONNREFUSED
        with mock.patch('socket.create_connection', side_effect=excp):
            target = targets.get_target(self.config_stack, args)
        self.assertEqual(target.name, targets.TARGET_QEMU)
        self.assertEqual(target.username, 'test_qemu_user')
        self.assertEqual(target.password, '56789')
        self.assertEqual(target.cpu, '3')
        self.assertEqual(target.ram, '4096')
        self.assertEqual(target.headless, False)
        self.assertEqual(target.port, 22220)
        self.assertEqual(target.overlay, '/tmp/cache/data/tmp/overlay.img')
        mock_isfile.assert_called_once_with('/tmp/disk.img')

    @mock.patch('ubuntu_system_tests.host.debian.get_dependencies',
                return_value=['dep1'])
    def test_get_target_ssh(self, mock_get_deps):
        self.set_config_values(
            device_password='1234', device_username='test_user',
            target_type=targets.TARGET_SSH, device_ip='test_device_ip',
            device_port='123456')
        args = DummyArgs()
        target = targets.get_target(self.config_stack, args)
        self.assertEqual(target.name, targets.TARGET_SSH)
        self.assertEqual(target.username, 'test_user')
        self.assertEqual(target.password, '1234')
        self.assertEqual(target.port, 123456)
        self.assertEqual(target.device_ip, 'test_device_ip')

    @mock.patch('ubuntu_system_tests.host.debian.get_dependencies',
                return_value=['dep1'])
    def test_get_target_ssh_with_custom_args(self, mock_get_deps):
        self.set_config_values(
            device_password='1234', device_username='test_user',
            target_type=targets.TARGET_SSH, device_ip='test_device_ip',
            device_port='123456')
        args = DummyArgs(port='9876', ip='testhost')
        target = targets.get_target(self.config_stack, args)
        self.assertEqual(target.name, targets.TARGET_SSH)
        self.assertEqual(target.username, 'test_user')
        self.assertEqual(target.password, '1234')
        self.assertEqual(target.port, 9876)
        self.assertEqual(target.device_ip, 'testhost')
