# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016-2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from unittest import mock

from ubuntu_system_tests.host import commands
from ubuntu_system_tests.common import config
from ubuntu_system_tests.selftests.utils import (
    check_any_item_in_list_contains_value,
    check_any_item_in_list_startswith_value,
    ConfigBaseTestCase,
    DummyArgs,
)
from ubuntu_system_tests.host.targets import get_target


class RunCommandTestCase(ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()
        self.args = DummyArgs(target='adb')

    def test_setup_network_command_contains_credentials(self):
        self.set_config_values()

        command = commands._get_device_network_command(
            'test_ssid', 'test_password')

        self.assertTrue(command.startswith(
            'nmcli device wifi connect \'test_ssid\' '
            'password \'test_password\''))


@mock.patch('ubuntu_system_tests.host.debian.get_dependencies',
            return_value=['dep1'])
@mock.patch('ubuntu_system_tests.common.ssh.get_ssh_config_commands')
class BuildAdtRunCommandTestCase():
    # TODO: extend the class from ConfigBaseTestCase so that test cases are run

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()
        self.args = DummyArgs(target='adb')
        with mock.patch(
                'ubuntu_system_tests.host.debian.get_dependencies') as mock_gd:
            mock_gd.return_value = ['dep1']
            self.target = get_target(self.config_stack, self.args)

    def test_adt_run_must_be_called_with_short_timeout(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()
        command = commands.build_adt_run_test_command(
            self.config_stack, self.args, self.target)
        command = ' '.join(command)
        separator = ' ' + commands._get_virt_separator()[0] + ' '
        run_arguments, _ = command.split(separator, maxsplit=1)
        self.assertIn('--timeout-short=600', run_arguments)
        mock_ssh_config_commands.assert_called_once_with()

    def test_adt_run_command_no_rw_file_mount(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()
        run_cmd = ' '.join(commands.build_adt_run_test_command(
            self.config_stack, self.args, self.target))
        self.assertNotIn('mount -o remount,rw /', run_cmd)
        mock_ssh_config_commands.assert_called_once_with()

    def test_adt_run_command_ro_file_mount(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()
        run_cmd = ' '.join(commands.build_adt_run_test_command(
            self.config_stack, self.args, self.target))
        self.assertIn('mount -o remount,ro /', run_cmd)
        mock_ssh_config_commands.assert_called_once_with()

    def test_adt_run_with_adb_serial(
            self, mock_ssh_config_commands, mock_get_dependencies):
        serial = 'test_serial'
        self.set_config_values(device_serial=serial)

        command = commands.build_adt_run_test_command(
            self.config_stack, self.args, self.target)
        command = ' '.join(command)

        _, adb_arguments = command.split('ssh -s adb --')
        self.assertIn('-s test_serial', adb_arguments)
        mock_ssh_config_commands.assert_called_once_with()

    def test_adt_run_without_adb_serial(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values(device_serial='')

        command = commands.build_adt_run_test_command(
            self.config_stack, self.args, self.target)
        command = ' '.join(command)

        _, adb_arguments = command.split('ssh -s adb --')
        self.assertNotIn('-s test_serial', adb_arguments)
        mock_ssh_config_commands.assert_called_once_with()

    def test_adt_run_command_with_network_ready_creates_no_network(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()
        self.args.network_ready = True
        command = commands.build_adt_run_test_command(
            self.config_stack, self.args, self.target)

        self.assertFalse(check_any_item_in_list_startswith_value(
            command, 'nmcli device wifi connect'))
        mock_ssh_config_commands.assert_called_once_with()

    def test_adt_run_command_with_default_parameters_creates_network(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()

        command = commands.build_adt_run_test_command(
            self.config_stack, self.args, self.target)

        self.assertTrue(check_any_item_in_list_startswith_value(
            command, 'nmcli device wifi connect'))
        mock_ssh_config_commands.assert_called_once_with()

    def test_enabling_verbose_option(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()
        self.args.verbose = True
        command = commands.build_adt_run_test_command(
            self.config_stack, self.args, self.target)
        self.assertTrue(check_any_item_in_list_contains_value(command, '-ddd'))
        mock_ssh_config_commands.assert_called_once_with()

    def test_verbose_disabled_by_default(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()
        command = commands.build_adt_run_test_command(
            self.config_stack, self.args, self.target)
        self.assertFalse(
            check_any_item_in_list_contains_value(command, '-ddd'))
        mock_ssh_config_commands.assert_called_once_with()


class TestAutopktestLegacyCommandOptions(ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()
        self.args = DummyArgs(target='adb')

    @mock.patch('ubuntu_system_tests.host.results.get_output_directory',
                return_value='/tmp/results')
    @mock.patch('ubuntu_system_tests.host.commands.get_autopkgtest_version',
                return_value=3.2)
    def test_adt_run_used_for_legacy(
            self, mock_get_version, mock_get_output_dir):
        self.set_config_values()
        commands.legacy_cmds = None
        cmd = commands._get_autopkgtest_cmd_start(self.config_stack, self.args)
        mock_get_version.assert_called_once_with()
        self.assertEqual(cmd[0], 'adt-run')
        self.assertTrue(mock_get_output_dir.called)

    @mock.patch('ubuntu_system_tests.host.results.get_output_directory',
                return_value='/tmp/results')
    @mock.patch('ubuntu_system_tests.host.commands.get_autopkgtest_version',
                return_value=4.1)
    def test_autopkgtest_used_for_current(
            self, mock_get_version, mock_get_output_dir):
        self.set_config_values()
        commands.legacy_cmds = None
        cmd = commands._get_autopkgtest_cmd_start(self.config_stack, self.args)
        mock_get_version.assert_called_once_with()
        self.assertEqual(cmd[0], 'autopkgtest')
        self.assertTrue(mock_get_output_dir.called)

    @mock.patch('ubuntu_system_tests.host.results.get_output_directory',
                return_value='/tmp/results')
    @mock.patch('ubuntu_system_tests.host.debian.get_dependencies',
                return_value=['dep1'])
    @mock.patch('ubuntu_system_tests.host.commands.get_autopkgtest_version',
                return_value=3.2)
    def test_unbuilt_tree_used_for_legacy(
            self, mock_get_version, mock_get_deps, mock_get_output_dir):
        self.set_config_values()
        commands.legacy_cmds = None
        cmd = commands.build_basic_adt_run_test_command(
            self.config_stack, DummyArgs(suites=[]), False)
        self.assertTrue(
            check_any_item_in_list_startswith_value(cmd, '--unbuilt-tree'))
        mock_get_version.assert_called_once_with()
        self.assertTrue(mock_get_output_dir.called)

    @mock.patch('ubuntu_system_tests.host.results.get_output_directory',
                return_value='/tmp/results')
    @mock.patch('ubuntu_system_tests.host.debian.get_dependencies',
                return_value=['dep1'])
    @mock.patch('ubuntu_system_tests.host.commands.get_autopkgtest_version',
                return_value=4.1)
    def test_unbuilt_tree_not_used_for_current(
            self, mock_get_version, mock_get_deps, mock_get_output_dir):
        self.set_config_values()
        commands.legacy_cmds = None
        cmd = commands.build_basic_adt_run_test_command(
            self.config_stack, DummyArgs(suites=[]), False)
        self.assertFalse(
            check_any_item_in_list_startswith_value(cmd, '--unbuilt-tree'))
        mock_get_version.assert_called_once_with()
        self.assertTrue(mock_get_output_dir.called)


class MockTarget:

    def __init__(self, **kwargs):
        for arg in kwargs.keys():
            setattr(self, arg, kwargs[arg])

    def snap_session(self):
        return self.snap


class TestUpgradeSetupCommands(ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.args = DummyArgs(upgrade='package1,package2')

    def test_get_upgrade_setup_command_snap(self):
        target = MockTarget(snap=True, config={'mount_fs_rw': False})
        cmd = commands._get_upgrade_setup_command(target, self.args.upgrade)
        self.assertEqual(
            cmd,
            ['--setup-commands',
             'snap refresh --devmode --edge package1; '
             'snap refresh --devmode --edge package2;'])

    def test_get_upgrade_setup_command_deb(self):
        target = MockTarget(snap=False, config={'mount_fs_rw': False})
        cmd = commands._get_upgrade_setup_command(target, self.args.upgrade)
        self.assertEqual(
            cmd,
            ['--setup-commands',
             'apt-get -y --no-install-recommends install package1 package2'])

    def test_get_upgrade_setup_includes_fs_rw(self):
        target = MockTarget(snap=False, config={'mount_fs_rw': True})
        cmd = commands._get_upgrade_setup_command(target, self.args.upgrade)
        self.assertIn('mount -o remount,rw /', cmd[1])
