# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from ubuntu_system_tests.common.config import get_device_config_stack
from ubuntu_system_tests.helpers.addressbook.app import AddressBook
from ubuntu_system_tests.helpers.calculator.app import Calculator
from ubuntu_system_tests.helpers.calculator.fixture_setup import (
    CalculatorConfigFixture,
)
from ubuntu_system_tests.helpers.calendar.app import Calendar
from ubuntu_system_tests.helpers.camera.app import Camera
from ubuntu_system_tests.helpers.clock.app import Clock
from ubuntu_system_tests.helpers.filemanager.app import FileManager
from ubuntu_system_tests.helpers.gallery.app import Gallery
from ubuntu_system_tests.helpers.terminal.app import Terminal
from ubuntu_system_tests.helpers.system_settings.app import (
    get_system_settings_app,
)
from ubuntu_system_tests.helpers.webbrowser.app import WebBrowser
from ubuntu_system_tests.tests.base import BaseUbuntuSystemTestCase

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class AppLaunchBase(BaseUbuntuSystemTestCase):

    def setUp(self):
        super().setUp()
        self.app = None
        self.config_stack = get_device_config_stack()

    def tearDown(self):
        if self.app:
            self.app.stop()
        super().tearDown()

    def check_app_launch(self, app_class, launch_count):
        self.app = app_class()
        if not self.app.is_installed():
            reason = '{n} not installed as {t}'.format(
                n=self.app.app_id, t=self.app.app_type)
            self.app = None
            self.skipTest(reason)
        for count in range(1, launch_count + 1):
            self.app.launch()
            self.assertIsNotNone(self.app.get_main_view())
            logger.info('App launch {c} successful'.format(c=count))
            if count < launch_count:
                # Only close the app here if this is not the last iteration.
                # Otherwise it will be closed automatically in tearDown.
                self.app.stop()


class LaunchAppOnceTestCase(AppLaunchBase):

    def test_launch_addressbook_app(self):
        self.check_app_launch(AddressBook, 1)

    def test_launch_camera_app(self):
        self.check_app_launch(Camera, 1)

    def test_launch_gallery_app(self):
        self.check_app_launch(Gallery, 1)

    def test_launch_calculator_app(self):
        self.useFixture(CalculatorConfigFixture(first_run=False))
        self.check_app_launch(Calculator, 1)

    def test_launch_calendar_app(self):
        self.check_app_launch(Calendar, 1)

    def test_launch_clock_app(self):
        self.check_app_launch(Clock, 1)

    def test_launch_filemanager_app(self):
        self.check_app_launch(FileManager, 1)

    def test_launch_terminal_app(self):
        self.check_app_launch(Terminal, 1)

    def test_launch_webbrowser_app(self):
        self.check_app_launch(WebBrowser, 1)

    def test_launch_systemsettings_app(self):
        app = get_system_settings_app(self.config_stack.get('unity8_mode'))
        self.check_app_launch(app, 1)


class LaunchAppTwiceTestCase(AppLaunchBase):

    def test_launch_addressbook_app_twice(self):
        self.check_app_launch(AddressBook, 2)

    def test_launch_camera_app_twice(self):
        self.check_app_launch(Camera, 2)

    def test_launch_gallery_app_twice(self):
        self.check_app_launch(Gallery, 2)

    def test_launch_calculator_app_twice(self):
        self.useFixture(CalculatorConfigFixture(first_run=False))
        self.check_app_launch(Calculator, 2)

    def test_launch_calendar_app_twice(self):
        self.check_app_launch(Calendar, 2)

    def test_launch_clock_app_twice(self):
        self.check_app_launch(Clock, 2)

    def test_launch_filemanager_app_twice(self):
        self.check_app_launch(FileManager, 2)

    def test_launch_terminal_app_twice(self):
        self.check_app_launch(Terminal, 2)

    def test_launch_webbrowser_app_twice(self):
        self.check_app_launch(WebBrowser, 2)

    def test_launch_systemsettings_app_twice(self):
        app = get_system_settings_app(self.config_stack.get('unity8_mode'))
        self.check_app_launch(app, 2)
